import asyncio
import logging

import regex

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.scrape as s_scrape
import synapse.lib.spooled as s_spooled
import synapse.lib.stormtypes as s_stormtypes

logger = logging.getLogger(__name__)

@s_stormtypes.registry.registerLib
class LibScrape(s_stormtypes.Lib):
    '''
    A Storm Library for providing helpers for scraping nodes from text.
    '''
    _storm_locals = (
        {'name': 'context', 'desc': '''
            Attempt to scrape information from a blob of text, getting the context information about the values found.

            Notes:
                This does call the ``scrape`` Storm interface if that behavior is enabled on the Cortex.

            Examples:
                Scrape some text and make nodes out of it::

                    for ($form, $valu, $info) in $lib.scrape.context($text) {
                        [ ( *$form ?= $valu ) ]
                    }
            ''',
         'type': {'type': 'function', '_funcname': '_methContext',
                  'args': (
                      {'name': 'text', 'type': 'str',
                       'desc': 'The text to scrape', },
                  ),
                  'returns': {'name': 'yields', 'type': 'dict',
                              'desc': 'A dictionary of scraped values, rule types, and offsets scraped from the text.',
                              }}},
        {'name': 'ndefs', 'desc': '''
            Attempt to scrape node form, value tuples from a blob of text.

            Examples:
                Scrape some text and attempt to make nodes out of it::

                    for ($form, $valu) in $lib.scrape($text) {
                        [ ( *$form ?= $valu ) ]
                    }''',
         'type': {'type': 'function', '_funcname': '_methNdefs',
                  'args': (
                      {'name': 'text', 'type': 'str',
                       'desc': 'The text to scrape', },
                  ),
                  'returns': {'name': 'yields', 'type': 'list',
                              'desc': 'A list of (form, value) tuples scraped from the text.', }}},
        {'name': 'genMatches', 'desc': '''
        genMatches is a generic helper function for constructing scrape interfaces using pure Storm.

        It accepts the text, a regex pattern, and produce results that can easily be used to create

        Notes:
            The pattern must have a named regular expression match for the key ``valu`` using the
            named group syntax. For example ``(somekey\\s)(?P<valu>[a-z0-9]+)\\s``.

        Examples:
            A scrape implementation with a regex that matches name keys in text::

                $re="(Name\\:\\s)(?P<valu>[a-z0-9]+)\\s"
                $form="ps:name"

                function scrape(text, form) {
                        $ret = $lib.list()
                        for ($valu, $info) in $lib.scrape.genMatches($text, $re) {
                            $ret.append(($form, $valu, $info))
                        }
                        return ( $ret )
                    }
        ''',
         'type': {'type': 'function', '_funcname': '_methGenMatches',
                  'args': (
                      {'name': 'text', 'type': 'str',
                       'desc': 'The text to scrape', },
                      {'name': 'pattern', 'type': 'str',
                       'desc': 'The regular expression pattern to match against.', },
                      {'name': 'fangs', 'type': 'list', 'default': None,
                       'desc': 'A list of (src, dst) pairs to refang from text. The src must be equal or larger '
                               'than the dst in length.'},
                      {'name': 'flags', 'type': 'int', 'default': regex.IGNORECASE,
                       'desc': 'Regex flags to use (defaults to IGNORECASE).'},
                  ),
                  'returns': {'name': 'yields', 'type': 'list',
                              'desc': ''}}}
    )
    _storm_lib_path = ('scrape', )

    def getObjLocals(self):
        return {
            'ndefs': self._methNdefs,
            'context': self._methContext,
            'genMatches': self._methGenMatches,
        }

    async def __call__(self, text, ptype=None, refang=True, unique=True):
        text = await s_stormtypes.tostr(text)
        form = await s_stormtypes.tostr(ptype, noneok=True)
        refang = await s_stormtypes.tobool(refang)
        unique = await s_stormtypes.tobool(unique)
        # Remove this in 3.0.0 since it is deprecated.
        s_common.deprecated('$lib.scrape()')
        await self.runt.warnonce('$lib.scrape() is deprecated. Use $lib.scrape.ndefs().')
        async with await s_spooled.Set.anit() as items:  # type: s_spooled.Set
            for item in s_scrape.scrape(text, ptype=form, refang=refang, first=False):
                if unique:
                    if item in items:
                        continue
                    await items.add(item)

                yield item
                await asyncio.sleep(0)

    @s_stormtypes.stormfunc(readonly=True)
    async def _methContext(self, text):
        text = await s_stormtypes.tostr(text)

        genr = self.runt.snap.view.scrapeIface(text)
        async for (form, valu, info) in genr:
            yield (form, valu, info)

    @s_stormtypes.stormfunc(readonly=True)
    async def _methNdefs(self, text):
        text = await s_stormtypes.tostr(text)

        genr = self.runt.snap.view.scrapeIface(text, unique=True)
        async for (form, valu, _) in genr:
            yield (form, valu)

    @s_stormtypes.stormfunc(readonly=True)
    async def _methGenMatches(self, text, pattern, fangs=None, flags=regex.IGNORECASE):
        text = await s_stormtypes.tostr(text)
        pattern = await s_stormtypes.tostr(pattern)
        fangs = await s_stormtypes.toprim(fangs)
        flags = await s_stormtypes.toint(flags)

        opts = {}
        regx = regex.compile(pattern, flags=flags)

        _fangs = None
        _fangre = None
        offsets = None
        scrape_text = text
        if fangs:
            _fangs = {src: dst for (src, dst) in fangs}
            _fangre = s_scrape.genFangRegex(_fangs)
            scrape_text, offsets = s_scrape.refang_text2(text, re=_fangre, fangs=_fangs)

        for info in s_scrape.genMatches(scrape_text, regx, opts=opts):
            valu = info.pop('valu')

            if _fangs and offsets:
                s_scrape._rewriteRawValu(text, offsets, info)

            yield valu, info



from __future__ import annotations

import os
import heapq
import bisect
import shutil
import asyncio
import logging
import contextlib

import regex

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.base as s_base
import synapse.lib.coro as s_coro
import synapse.lib.slabseqn as s_slabseqn
import synapse.lib.lmdbslab as s_lmdbslab

from typing import List, Tuple, Dict, Optional, Any, AsyncIterator

logger = logging.getLogger(__name__)

seqnslabre = regex.compile(r'^seqn([0-9a-f]{16})\.lmdb$')

class MultiSlabSeqn(s_base.Base):
    '''
    An append-optimized sequence of byte blobs stored across multiple slabs for fast rotating/culling
    '''

    async def __anit__(self,  # type: ignore
                       dirn: str,
                       opts: Optional[Dict] = None,
                       slabopts: Optional[Dict] = None):
        '''
        Args:
            dirn (str):  directory where to store the slabs
            opts (Optional[Dict]):  options for this multislab
            slabopts (Optional[Dict]):  options to pass through to the slab creation

        '''

        await s_base.Base.__anit__(self)

        if opts is None:
            opts = {}

        self.offsevents: List[Tuple[int, int, asyncio.Event]] = []  # as a heap
        self._waitcounter = 0

        self.dirn: str = dirn
        s_common.gendir(self.dirn)
        self.slabopts: Dict[str, Any] = {} if slabopts is None else slabopts

        # The last/current slab
        self.tailslab: Optional[s_lmdbslab.Slab] = None
        self.tailseqn: Optional[s_slabseqn.SlabSeqn] = None

        # The most recently accessed slab/seqn that isn't the tail
        self._cacheslab: Optional[s_lmdbslab.Slab] = None
        self._cacheseqn: Optional[s_slabseqn.SlabSeqn] = None
        self._cacheridx: Optional[int] = None

        # A startidx -> (Slab, Seqn) dict for all open Slabs, so we don't accidentally open the same Slab twice
        self._openslabs: Dict[int, Tuple[s_lmdbslab.Slab, s_slabseqn.SlabSeqn]] = {}

        # Lock to avoid an open race
        self._openlock = asyncio.Lock()

        await self._discoverRanges()

        async def fini():
            for slab, _ in list(self._openslabs.values()):
                # We incref the slabs, so might have to fini multiple times
                count = 1
                while count:
                    count = await slab.fini()

        self.onfini(fini)

    def __repr__(self):
        return f'MultiSlabSeqn: {self.dirn!r}'

    @staticmethod
    def _getFirstIndx(slab) -> Optional[int]:
        db = slab.initdb('info')
        bytz = slab.get(b'firstindx', db=db)
        if bytz is None:
            return 0
        return s_common.int64un(bytz)

    @staticmethod
    def _setFirstIndx(slab, indx) -> bool:
        db = slab.initdb('info')
        return slab.put(b'firstindx', s_common.int64en(indx), db=db)

    async def _discoverRanges(self):
        '''
        Go through the slabs and get the starting indices of the sequence in each slab
        '''
        fnstartidx = 0
        lastidx = None
        self._ranges: List[int] = []  # Starting offsets of all the slabs in order
        self.firstindx = 0  # persistently-stored indicator of lowest index
        self.indx = 0  # The next place an add() will go
        lowindx = None

        # Make sure the files are in order

        for fn in sorted(s_common.listdir(self.dirn, glob='*seqn' + '[abcdef01234567890]' * 16 + '.lmdb')):

            if not os.path.isdir(fn):
                logger.warning(f'Found a non-directory {fn} where a directory should be')
                continue
            match = seqnslabre.match(os.path.basename(fn))
            assert match

            newstartidx = int(match.group(1), 16)

            assert newstartidx >= fnstartidx

            fnstartidx = newstartidx

            if lowindx is None:
                lowindx = fnstartidx

            if lastidx is not None:
                if fnstartidx <= lastidx:
                    mesg = f'Multislab:  overlapping files ({fn}).  Previous last index is {lastidx}.'
                    raise s_exc.BadCoreStore(mesg=mesg)

                if fnstartidx != lastidx + 1:
                    logger.debug(f'Multislab:  gap in indices at {fn}.  Previous last index is {lastidx}.')

            async with await s_lmdbslab.Slab.anit(fn, **self.slabopts) as slab:
                self.firstindx = self._getFirstIndx(slab)
                # We use the old name of the sequence to ease migration from the old system
                seqn = slab.getSeqn('nexuslog')

                firstitem = seqn.first()

                if firstitem is None:
                    self.indx = fnstartidx
                else:
                    self.indx = seqn.indx

                    firstidx = firstitem[0]  # might not match the separately stored first index due to culling

                    if firstidx < fnstartidx:
                        raise s_exc.BadCoreStore('Multislab:  filename inconsistent with contents')

                    lastidx = seqn.index() - 1

            self._ranges.append(fnstartidx)

        # An admin might have manually culled by rm'ing old slabs.  Update firstidx accordingly.
        if lowindx is not None and lowindx > self.firstindx:
            self.firstindx = lowindx

        if self.firstindx > self.indx:
            raise s_exc.BadCoreStore('Invalid firstindx value')

        await self._initTailSlab(fnstartidx)

    @staticmethod
    def slabFilename(dirn: str, indx: int):
        return s_common.genpath(dirn, f'seqn{indx:016x}.lmdb')

    async def _initTailSlab(self, indx: int) -> int:
        if self.tailslab:
            await self.tailslab.fini()

        self.tailslab, self.tailseqn = await self._makeSlab(indx)

        if not self.tailslab.dbexists('info'):
            self._setFirstIndx(self.tailslab, self.firstindx)
            self.tailseqn.indx = indx
            self._ranges.append(indx)

        return indx

    def _wake_waiters(self) -> None:
        while self.offsevents and self.offsevents[0][0] < self.indx:
            _, _, evnt = heapq.heappop(self.offsevents)
            evnt.set()

    async def rotate(self) -> int:
        '''
        Rotate the Nexus log at the current index.

        Note:
            After this executes the tailseqn will be empty.
            Waiting for this indx to be written will indicate
            when it is possible to cull 1 minus the return value
            such that the rotated seqn is deleted.

        Returns:
            int: The starting index of the new seqn
        '''
        assert self.tailslab and self.tailseqn and self._ranges

        if self.indx <= self._ranges[-1]:
            logger.info('Seqn %s at indx %d is empty', self.tailslab.path, self.indx)
            return self._ranges[-1]

        logger.info('Rotating %s at indx %d', self.tailslab.path, self.indx)
        return await self._initTailSlab(self.indx)

    async def cull(self, offs: int) -> bool:
        '''
        Remove entries up to (and including) the given offset.
        '''

        logger.info('Culling %s at offs %d', self.dirn, offs)

        # Note:  we don't bother deleting the rows from inside a partially culled slab.  We just update self.firstindx
        # so nothing will return those rows anymore.  We only delete from disk entire slabs once they are culled.

        if offs < self.firstindx:
            logger.warning('Unable to cull %s; offs (%d) < starting indx (%d)', self.dirn, offs, self.firstindx)
            return False

        # We keep at least one entry;  this avoids offsets possibly going lower after a restart
        if offs >= self.indx - 1:
            logger.warning('Unable to cull %s at offs %d; must keep at least one entry', self.dirn, offs)
            return False

        if self._cacheridx is not None:
            self._cacheridx = None
            assert self._cacheslab
            await self._cacheslab.fini()
            self._cacheslab = self._cacheseqn = None

        del_ridx = None
        for ridx in range(len(self._ranges) - 1):
            startidx = self._ranges[ridx]

            if self._openslabs.get(startidx):
                raise s_exc.SlabInUse(mesg='Attempt to cull while another task is still using it')

            fn = self.slabFilename(self.dirn, startidx)
            if offs < self._ranges[ridx + 1] - 1:
                logger.warning('Log %s will not be deleted since offs is less than last indx', fn)
                break

            optspath = s_common.switchext(fn, ext='.opts.yaml')
            try:
                os.unlink(optspath)
            except FileNotFoundError:  # pragma: no cover
                pass

            logger.info('Removing log %s with startidx %d', fn, startidx)
            shutil.rmtree(fn)
            del_ridx = ridx

            await asyncio.sleep(0)

        self.firstindx = offs + 1
        self._setFirstIndx(self.tailslab, offs + 1)

        if del_ridx is not None:
            del self._ranges[:del_ridx + 1]

        # Log if there was an attempt to cull into the tailseqn
        if offs >= self._ranges[-1]:
            fn = self.tailslab.path
            logger.warning('Log %s will not be deleted since offs is in the currently active log', fn)

        return True

    async def _makeSlab(self, startidx: int) -> Tuple[s_lmdbslab.Slab, s_slabseqn.SlabSeqn]:

        async with self._openlock:  # Avoid race in two tasks making the same slab

            item = self._openslabs.get(startidx)
            if item is not None:
                item[0].incref()
                return item

            fn = self.slabFilename(self.dirn, startidx)

            slab = await s_lmdbslab.Slab.anit(fn, **self.slabopts)
            seqn = slab.getSeqn('nexuslog')

            self._openslabs[startidx] = slab, seqn

            def fini():
                self._openslabs.pop(startidx, None)

            slab.onfini(fini)

            return slab, seqn

    @contextlib.asynccontextmanager
    async def _getSeqn(self, ridx: int) -> AsyncIterator[s_slabseqn.SlabSeqn]:
        '''
        Get the sequence corresponding to an index into self._ranges
        '''
        if ridx == len(self._ranges) - 1:
            assert self.tailslab and self.tailseqn
            slab, seqn = self.tailslab, self.tailseqn

        elif ridx == self._cacheridx:
            assert self._cacheslab and self._cacheseqn
            slab, seqn = self._cacheslab, self._cacheseqn

        else:
            startidx = self._ranges[ridx]

            self._cacheridx = None

            if self._cacheslab is not None:
                await self._cacheslab.fini()

            slab, seqn = self._cacheslab, self._cacheseqn = await self._makeSlab(startidx)
            self._cacheridx = ridx

        slab.incref()
        try:
            yield seqn
        finally:
            await slab.fini()

    async def add(self, item: Any, indx=None) -> int:
        '''
        Add a single item to the sequence.
        '''
        advances = True

        if indx is not None:
            if indx < self.firstindx:
                raise s_exc.BadIndxValu(mesg=f'indx lower than first index in sequence {self.firstindx}')

            if indx < self._ranges[-1]:
                ridx = self._getRangeIndx(indx)
                assert ridx is not None

                async with self._getSeqn(ridx) as seqn:
                    seqn.add(item, indx=indx)

                return indx

            if indx >= self.indx:
                self.indx = indx
            else:
                advances = False
        else:
            indx = self.indx

        assert self.tailseqn
        retn = self.tailseqn.add(item, indx=indx)

        if advances:
            self.indx += 1

            self._wake_waiters()

        return retn

    async def last(self) -> Optional[Tuple[int, Any]]:
        ridx = self._getRangeIndx(self.indx - 1)
        if ridx is None:
            return None

        async with self._getSeqn(ridx) as seqn:
            return seqn.last()

    def index(self) -> int:
        '''
        Return the current index to be used
        '''
        return self.indx

    def setIndex(self, indx: int) -> None:
        self.indx = indx

    def _getRangeIndx(self, offs: int) -> Optional[int]:
        '''
        Return the index into self._ranges that contains the offset
        '''
        if offs < self.firstindx:
            return None

        indx = bisect.bisect_right(self._ranges, offs)
        assert indx

        return indx - 1

    async def iter(self, offs: int) -> AsyncIterator[Tuple[int, Any]]:
        '''
        Iterate over items in a sequence from a given offset.

        Args:
            offs (int): The offset to begin iterating from.

        Yields:
            (indx, valu): The index and valu of the item.
        '''
        offs = max(offs, self.firstindx)

        ri = ridx = self._getRangeIndx(offs)
        assert ridx is not None

        # ranges could get appended while iterating due to a rotation
        while ri < len(self._ranges):
            if ri > ridx:
                offs = self._ranges[ri]

            async with self._getSeqn(ri) as seqn:
                for item in seqn.iter(offs):
                    yield item

            ri += 1

    async def gets(self, offs, wait=True) -> AsyncIterator[Tuple[int, Any]]:
        '''
        Just like iter, but optionally waits for new entries once the end is reached.
        '''
        while True:

            async for (indx, valu) in self.iter(offs):
                yield (indx, valu)
                offs = indx + 1

            if not wait:
                return

            await self.waitForOffset(self.indx)

    async def get(self, offs: int) -> Any:
        '''
        Retrieve a single row by offset
        '''
        ridx = self._getRangeIndx(offs)
        if ridx is None:
            raise s_exc.BadIndxValu(mesg=f'offs lower than first index {self.firstindx}')

        async with self._getSeqn(ridx) as seqn:
            return seqn.get(offs)

    def getOffsetEvent(self, offs: int) -> asyncio.Event:
        '''
        Returns an asyncio Event that will be set when the particular offset is written.  The event will be set if the
        offset has already been reached.
        '''
        evnt = asyncio.Event()

        if offs < self.indx:
            evnt.set()
            return evnt

        # We add a simple counter to the tuple to cause stable (and FIFO) sorting and prevent ties
        heapq.heappush(self.offsevents, (offs, self._waitcounter, evnt))

        self._waitcounter += 1

        return evnt

    async def waitForOffset(self, offs: int, timeout=None) -> bool:
        '''
        Returns:
            true if the event got set, False if timed out
        '''

        if offs < self.indx:
            return True

        evnt = self.getOffsetEvent(offs)
        return await s_coro.event_wait(evnt, timeout=timeout)



import unittest.mock as mock

import synapse.common as s_common

import synapse.lib.cmdr as s_cmdr

import synapse.tests.utils as s_t_utils

class CmdTriggersTest(s_t_utils.SynTest):

    async def test_triggers(self):

        async with self.getTestCoreAndProxy() as (realcore, core):

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)

            await cmdr.runCmdLine('trigger list')
            self.true(outp.expect('No triggers found'))

            await cmdr.runCmdLine('trigger add node:add test:str {[ test:int=1 ] }')
            trigs = await realcore.view.listTriggers()
            self.len(1, trigs)
            self.true(outp.expect(f'Added trigger {trigs[0][0]}'))
            await s_common.aspin(core.eval('[ test:str=foo ]'))
            await self.agenlen(1, core.eval('test:int'))

            await cmdr.runCmdLine('trigger add tag:add test:str #footag.* {[ +#count test:str=$tag ]}')
            await s_common.aspin(core.eval('[ test:str=bar +#footag.bar ]'))
            await self.agenlen(1, core.eval('#count'))
            await self.agenlen(1, core.eval('test:str=footag.bar'))

            await cmdr.runCmdLine('trigger add prop:set --disabled test:type10:intprop {[ test:int=6 ]}')
            outp.clear()
            await cmdr.runCmdLine('trigger list')
            self.true(outp.expect('user'))
            self.true(outp.expect('root'))
            goodbuid = outp.mesgs[-2].split()[1][:6]
            goodbuid2 = outp.mesgs[-1].split()[1][:6]

            # Trigger is created disabled, so no nodes yet
            await self.agenlen(0, core.eval('test:int=6'))
            await cmdr.runCmdLine(f'trigger enable {goodbuid2}')

            # Trigger is enabled, so it should fire
            await s_common.aspin(core.eval('[ test:type10=1 :intprop=25 ]'))
            await self.agenlen(1, core.eval('test:int=6'))

            await cmdr.runCmdLine(f'trigger del {goodbuid}')
            self.true(outp.expect('Deleted trigger'))

            await cmdr.runCmdLine(f'trigger del deadbeef12341234')
            self.true(outp.expect('does not match'))

            outp.clear()
            await cmdr.runCmdLine(f'trigger enable deadbeef12341234')
            self.true(outp.expect('does not match'))

            outp.clear()
            await cmdr.runCmdLine(f'trigger disable deadbeef12341234')
            self.true(outp.expect('does not match'))

            await cmdr.runCmdLine(f'trigger disable {goodbuid2}')
            self.true(outp.expect('Disabled trigger'))

            await cmdr.runCmdLine(f'trigger enable {goodbuid2}')
            self.true(outp.expect('Enabled trigger'))

            await cmdr.runCmdLine(f'trigger mod {goodbuid2} {{[ test:str=different ]}}')
            self.true(outp.expect('Modified trigger'))

            outp.clear()
            await cmdr.runCmdLine('trigger mod deadbeef12341234 {#foo}')
            self.true(outp.expect('does not match'))

            await cmdr.runCmdLine('trigger add tag:add #another {[ +#count2 ]}')

            # Syntax mistakes
            await cmdr.runCmdLine(f'trigger')
            self.true(outp.expect('Manipulate triggers in a '))

            await cmdr.runCmdLine('trigger add tag:add another {[ +#count2 ]}')
            self.true(outp.expect('Missing tag parameter'))

            await cmdr.runCmdLine('trigger add tug:udd another {[ +#count2 ]}')
            self.true(outp.expect('invalid choice'))

            await cmdr.runCmdLine('trigger add tag:add')
            self.true(outp.expect('trigger add: error: the following'))

            await cmdr.runCmdLine('trigger add tag:add inet:ipv4')
            self.true(outp.expect('Missing argument for trigger add'))

            await cmdr.runCmdLine('trigger add')
            self.true(outp.expect('Add triggers in a cortex.'))

            await cmdr.runCmdLine('trigger add tag:add #foo #bar')
            self.true(outp.expect('single tag'))

            await cmdr.runCmdLine('trigger add tag:add {test:str} {test:str}')
            self.true(outp.expect('single query'))

            await cmdr.runCmdLine('trigger add node:add test:str #foo {test:str}')
            self.true(outp.expect('node:* does not support'))

            await cmdr.runCmdLine('trigger add prop:set #foo {test:str}')
            self.true(outp.expect('Missing prop parameter'))

            await cmdr.runCmdLine('trigger add prop:set test:type10.intprop #foo {test:str}')
            self.true(outp.expect('prop:set does not support a tag'))

            await cmdr.runCmdLine('trigger add node:add test:str test:int {test:str}')
            self.true(outp.expect('Only a single form'))

            await cmdr.runCmdLine('trigger add prop:set test:type10.intprop test:str {test:str}')
            self.true(outp.expect('single prop'))

            await cmdr.runCmdLine('trigger add tag:add #tag test:int')
            self.true(outp.expect('Missing query'))

            await cmdr.runCmdLine('trigger add node:add #tag1 {test:str}')
            self.true(outp.expect('Missing form'))

            await cmdr.runCmdLine(f'trigger mod {goodbuid2} test:str')
            self.true(outp.expect('start with {'))

            # Bad storm syntax
            await cmdr.runCmdLine('trigger add node:add test:str {[ | | test:int=1 ] }')
            self.true(outp.expect('BadSyntax'))

            # (Regression) Just a command as the storm query
            await cmdr.runCmdLine('trigger add Node:add test:str {[ test:int=99 ] | spin }')
            await s_common.aspin(core.eval('[ test:str=foo4 ]'))
            await self.agenlen(1, core.eval('test:int=99'))

            # Test manipulating triggers as another user
            bond = await realcore.auth.addUser('bond')

            async with realcore.getLocalProxy(user='bond') as tcore:

                toutp = self.getTestOutp()
                tcmdr = await s_cmdr.getItemCmdr(tcore, outp=toutp)

                await tcmdr.runCmdLine('trigger list')
                self.true(toutp.expect('No triggers found'))

                await tcmdr.runCmdLine(f'trigger mod {goodbuid2} {{[ test:str=yep ]}}')
                self.true(toutp.expect('provided iden does not match'))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger disable {goodbuid2}')
                self.true(toutp.expect('provided iden does not match'))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger enable {goodbuid2}')
                self.true(toutp.expect('provided iden does not match'))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger del {goodbuid2}')
                self.true(toutp.expect('provided iden does not match'))

                # Give explicit perm
                await core.addUserRule(bond.iden, (True, ('trigger', 'get')))

                toutp.clear()
                await tcmdr.runCmdLine('trigger list')
                self.true(toutp.expect('root'))

                await core.addUserRule(bond.iden, (True, ('trigger', 'set')))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger mod {goodbuid2} {{[ test:str=yep ]}}')
                self.true(toutp.expect('Modified trigger'))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger disable {goodbuid2}')
                self.true(toutp.expect('Disabled trigger '))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger enable {goodbuid2}')
                self.true(toutp.expect('Enabled trigger '))

                await core.addUserRule(bond.iden, (True, ('trigger', 'del')))

                toutp.clear()
                await tcmdr.runCmdLine(f'trigger del {goodbuid2}')
                self.true(toutp.expect('Deleted trigger '))



import synapse.vendor.utils as s_v_utils

from synapse.vendor.xrpl.constants import CryptoAlgorithm
from synapse.vendor.xrpl.core import addresscodec


class TestCodec(s_v_utils.VendorTest):
    # seed tests

    def test_seed_encode_decode_secp256k1(self):
        hex_string = "CF2DE378FBDD7E2EE87D486DFB5A7BFF"
        encoded_string = "sn259rEFXrQrWyx3Q7XneWcwV6dfL"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_seed(
            hex_string_bytes, CryptoAlgorithm.SECP256K1
        )
        self.assertEqual(encode_result, encoded_string)

        decode_result, encoding_type = addresscodec.decode_seed(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)
        self.assertEqual(encoding_type, CryptoAlgorithm.SECP256K1)

    def test_seed_encode_decode_secp256k1_low(self):
        hex_string = "00000000000000000000000000000000"
        encoded_string = "sp6JS7f14BuwFY8Mw6bTtLKWauoUs"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_seed(
            hex_string_bytes, CryptoAlgorithm.SECP256K1
        )
        self.assertEqual(encode_result, encoded_string)

        decode_result, encoding_type = addresscodec.decode_seed(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)
        self.assertEqual(encoding_type, CryptoAlgorithm.SECP256K1)

    def test_seed_encode_decode_secp256k1_high(self):
        hex_string = "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
        encoded_string = "saGwBRReqUNKuWNLpUAq8i8NkXEPN"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_seed(
            hex_string_bytes, CryptoAlgorithm.SECP256K1
        )
        self.assertEqual(encode_result, encoded_string)

        decode_result, encoding_type = addresscodec.decode_seed(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)
        self.assertEqual(encoding_type, CryptoAlgorithm.SECP256K1)

    def test_seed_encode_decode_ed25519(self):
        hex_string = "4C3A1D213FBDFB14C7C28D609469B341"
        encoded_string = "sEdTM1uX8pu2do5XvTnutH6HsouMaM2"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_seed(
            hex_string_bytes, CryptoAlgorithm.ED25519
        )
        self.assertEqual(encode_result, encoded_string)

        decode_result, encoding_type = addresscodec.decode_seed(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)
        self.assertEqual(encoding_type, CryptoAlgorithm.ED25519)

    def test_seed_encode_decode_ed25519_low(self):
        hex_string = "00000000000000000000000000000000"
        encoded_string = "sEdSJHS4oiAdz7w2X2ni1gFiqtbJHqE"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_seed(
            hex_string_bytes, CryptoAlgorithm.ED25519
        )
        self.assertEqual(encode_result, encoded_string)

        decode_result, encoding_type = addresscodec.decode_seed(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)
        self.assertEqual(encoding_type, CryptoAlgorithm.ED25519)

    def test_seed_encode_decode_ed25519_high(self):
        hex_string = "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
        encoded_string = "sEdV19BLfeQeKdEXyYA4NhjPJe6XBfG"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_seed(
            hex_string_bytes, CryptoAlgorithm.ED25519
        )
        self.assertEqual(encode_result, encoded_string)

        decode_result, encoding_type = addresscodec.decode_seed(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)
        self.assertEqual(encoding_type, CryptoAlgorithm.ED25519)

    def test_seed_encode_decode_too_small(self):
        hex_string = "CF2DE378FBDD7E2EE87D486DFB5A7B"
        hex_string_bytes = bytes.fromhex(hex_string)

        self.assertRaises(
            addresscodec.XRPLAddressCodecException,
            addresscodec.encode_seed,
            hex_string_bytes,
            CryptoAlgorithm.SECP256K1,
        )

    def test_seed_encode_decode_too_big(self):
        hex_string = "CF2DE378FBDD7E2EE87D486DFB5A7BFFFF"
        hex_string_bytes = bytes.fromhex(hex_string)

        self.assertRaises(
            addresscodec.XRPLAddressCodecException,
            addresscodec.encode_seed,
            hex_string_bytes,
            CryptoAlgorithm.SECP256K1,
        )

    # classic_address tests

    def test_classic_address_encode_decode(self):
        hex_string = "BA8E78626EE42C41B46D46C3048DF3A1C3C87072"
        encoded_string = "rJrRMgiRgrU6hDF4pgu5DXQdWyPbY35ErN"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_classic_address(hex_string_bytes)
        self.assertEqual(encode_result, encoded_string)

        decode_result = addresscodec.decode_classic_address(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)

    def test_encode_classic_address_bad_length(self):
        hex_string = "ABCDEF"
        hex_string_bytes = bytes.fromhex(hex_string)

        self.assertRaises(
            addresscodec.XRPLAddressCodecException,
            addresscodec.encode_classic_address,
            hex_string_bytes,
        )

    # node_public_key test

    def test_node_public_key_encode_decode(self):
        hex_string = (
            "0388E5BA87A000CB807240DF8C848EB0B5FFA5C8E5A521BC8E105C0F0A44217828"
        )
        encoded_string = "n9MXXueo837zYH36DvMc13BwHcqtfAWNJY5czWVbp7uYTj7x17TH"
        hex_string_bytes = bytes.fromhex(hex_string)

        encode_result = addresscodec.encode_node_public_key(hex_string_bytes)
        self.assertEqual(encode_result, encoded_string)

        decode_result = addresscodec.decode_node_public_key(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)

    # account public key test

    def test_encode_account_public_key(self):
        hex_string = (
            "023693F15967AE357D0327974AD46FE3C127113B1110D6044FD41E723689F81CC6"
        )
        encoded_string = "aB44YfzW24VDEJQ2UuLPV2PvqcPCSoLnL7y5M1EzhdW4LnK5xMS3"
        hex_string_bytes = bytes.fromhex(hex_string)

        result = addresscodec.encode_account_public_key(hex_string_bytes)
        self.assertEqual(result, encoded_string)

        encode_result = addresscodec.encode_account_public_key(hex_string_bytes)
        self.assertEqual(encode_result, encoded_string)

        decode_result = addresscodec.decode_account_public_key(encoded_string)
        self.assertEqual(decode_result, hex_string_bytes)



"""High-level XRPL exports."""
#
# Copyright (c) 2021, XRP Ledger Foundation
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
# FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
# OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# -----------------------------------------------------------------------------------------------------------------------0
#
# The vendored library is a subset of the xrpl-py v1.1.1 library for the purposes of Cryptocurrency address validation.
# It has been modified for vendored imports.
# '''

from synapse.vendor.xrpl import core
from synapse.vendor.xrpl.constants import CryptoAlgorithm, XRPLException

__all__ = [
    "CryptoAlgorithm",
    "XRPLException",
    "core",
]



import os
import synapse.tests.utils as s_test

import synapse.common as s_common
import synapse.lib.output as s_output
import synapse.lib.msgpack as s_msgpack
import synapse.tools.storm as s_t_storm

class StormCliTest(s_test.SynTest):

    async def test_tools_storm(self):

        async with self.getTestCore() as core:

            await core.addTagProp('foo', ('int', {}), {})

            pars = s_t_storm.getArgParser()
            opts = pars.parse_args(('woot',))
            self.eq('woot', opts.cortex)

            q = '$lib.model.ext.addFormProp(inet:ipv4, "_test:score", (int, $lib.dict()), $lib.dict())'
            await core.callStorm(q)

            async with core.getLocalProxy() as proxy:

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('[inet:ipv4=1.2.3.4 +#foo=2012 +#bar +#baz:foo=10 :_test:score=7]')
                    text = str(outp)
                    self.isin('.....', text)
                    self.isin('inet:ipv4=1.2.3.4', text)
                    self.isin(':type = unicast', text)
                    self.isin(':_test:score = 7', text)
                    self.isin('.created = ', text)
                    self.isin('#bar', text)
                    self.isin('#baz:foo = 10', text)
                    self.isin('#foo = (2012/01/01 00:00:00.000, 2012/01/01 00:00:00.001)', text)
                    self.isin('complete. 1 nodes in', text)

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('!quit')
                    self.isin('o/', str(outp))
                    self.true(scli.isfini)

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('!help')
                    self.isin('!quit', str(outp))

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('$lib.print(woot)')
                    self.isin('woot', str(outp))

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('$lib.warn(woot)')
                    self.isin('WARNING: woot', str(outp))

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('---')
                    self.isin("---\n ^\nSyntax Error: No terminal defined for '-' at line 1 col 2", str(outp))

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('spin |' + ' ' * 80 + '---')
                    self.isin("...                             ---\n                                 ^", str(outp))

                outp = s_output.OutPutStr()
                async with await s_t_storm.StormCli.anit(proxy, outp=outp) as scli:
                    await scli.runCmdLine('---' + ' ' * 80 + 'spin')
                    self.isin("---                            ...\n ^", str(outp))

            lurl = core.getLocalUrl()

            outp = s_output.OutPutStr()
            await s_t_storm.main((lurl, '$lib.print(woot)'), outp=outp)
            self.isin('woot', str(outp))

            outp = s_output.OutPutStr()
            await s_t_storm.main((lurl, f'!runfile --help'), outp=outp)
            self.isin('Run a local storm file', str(outp))

            with self.getTestDir() as dirn:

                path = os.path.join(dirn, 'foo.storm')
                with open(path, 'wb') as fd:
                    fd.write(b'$lib.print(woot)')

                outp = s_output.OutPutStr()
                await s_t_storm.main((lurl, f'!runfile {path}'), outp=outp)
                self.isin(f'running storm file: {path}', str(outp))
                self.isin('woot', str(outp))

                outp = s_output.OutPutStr()
                await s_t_storm.main((lurl, f'!runfile /newp.storm'), outp=outp)
                self.isin(f'no such file: /newp.storm', str(outp))

                outp = s_output.OutPutStr()
                await s_t_storm.main((lurl, f'!pushfile /newp'), outp=outp)
                self.isin(f'no such file: /newp', str(outp))

                outp = s_output.OutPutStr()
                await s_t_storm.main((lurl, f'!pushfile {path}'), outp=outp)
                text = str(outp)
                self.isin(f'uploading file: {path}', text)
                self.isin(':name = foo.storm', text)
                self.isin(':sha256 = c00adfcc316f8b00772cdbce2505b9ea539d74f42861801eceb1017a44344ed3', text)

                outp = s_output.OutPutStr()
                path = os.path.join(dirn, 'bar.storm')
                await s_t_storm.main((lurl, f'!pullfile c00adfcc316f8b00772cdbce2505b9ea539d74f42861801eceb1017a44344ed3 {path}'), outp=outp)

                text = str(outp)
                self.isin('downloading sha256: c00adfcc316f8b00772cdbce2505b9ea539d74f42861801eceb1017a44344ed3', text)
                self.isin(f'saved to: {path}', text)

                with s_common.genfile(path) as fd:
                    self.isin('woot', fd.read().decode())

                outp = s_output.OutPutStr()
                await s_t_storm.main((lurl, f'!pullfile c11adfcc316f8b00772cdbce2505b9ea539d74f42861801eceb1017a44344ed3 {path}'), outp=outp)
                text = str(outp)
                self.isin('Axon does not contain the requested file.', text)

                await scli.runCmdLine('[test:str=foo +#foo +#bar +#baz]')
                await scli.runCmdLine('[test:str=bar +#foo +#bar +#baz]')

                path = os.path.join(dirn, 'export1.nodes')
                await s_t_storm.main((lurl, f'!export {path} {{ test:str }}'), outp=outp)
                text = str(outp)
                self.isin(f'saved 2 nodes to: {path}', text)

                with open(path, 'rb') as fd:
                    byts = fd.read()
                    podes = [i[1] for i in s_msgpack.Unpk().feed(byts)]
                    self.sorteq(('bar', 'foo'), [p[0][1] for p in podes])
                    for pode in podes:
                        self.sorteq(('bar', 'baz', 'foo'), pode[1]['tags'])

                path = os.path.join(dirn, 'export2.nodes')
                q = f'!export {path} {{ test:str }} --include-tags foo bar'
                await s_t_storm.main((lurl, q), outp=outp)
                text = str(outp)
                self.isin(f'saved 2 nodes to: {path}', text)

                with open(path, 'rb') as fd:
                    byts = fd.read()
                    podes = [i[1] for i in s_msgpack.Unpk().feed(byts)]
                    self.sorteq(('bar', 'foo'), [p[0][1] for p in podes])
                    for pode in podes:
                        self.sorteq(('bar', 'foo'), pode[1]['tags'])

                path = os.path.join(dirn, 'export3.nodes')
                q = f'!export {path} {{ test:str }} --no-tags'
                await s_t_storm.main((lurl, q), outp=outp)
                text = str(outp)
                self.isin(f'saved 2 nodes to: {path}', text)

                with open(path, 'rb') as fd:
                    byts = fd.read()
                    podes = [i[1] for i in s_msgpack.Unpk().feed(byts)]
                    self.sorteq(('bar', 'foo'), [p[0][1] for p in podes])
                    for pode in podes:
                        self.eq({}, pode[1]['tags'])

                await s_t_storm.main((lurl, f'!export {path} {{ test:newp }}'), outp=outp)
                text = str(outp)
                self.isin('No property named test:newp.', text)



import json
import uuid
import asyncio
import logging
import datetime

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.lib.coro as s_coro
import synapse.lib.node as s_node
import synapse.lib.stormctrl as s_stormctrl
import synapse.lib.stormtypes as s_stormtypes

import stix2validator

logger = logging.getLogger(__name__)

def uuid5(valu=None):
    guid = s_common.guid(valu=valu)
    return str(uuid.UUID(guid, version=5))

def uuid4(valu=None):
    guid = s_common.guid(valu=valu)
    return str(uuid.UUID(guid, version=4))

SYN_STIX_EXTENSION_ID = f'extension-definition--{uuid4("bf1c0a4d90ee557ac05055385971f17c")}'

_DefaultConfig = {

    'maxsize': 10000,

    'forms': {

        'ou:campaign': {
            'default': 'campaign',
            'stix': {
                'campaign': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'description': '+:desc return(:desc)',
                        'objective': '+:goal :goal -> ou:goal +:name return(:name)',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                    },
                    'rels': (
                        ('attributed-to', 'threat-actor', ':org -> ou:org'),
                        ('originates-from', 'location', ':org -> ou:org :hq -> geo:place'),
                        ('targets', 'identity', '-> risk:attack :target:org -> ou:org'),
                        ('targets', 'identity', '-> risk:attack :target:person -> ps:person'),
                        ('targets', 'vulnerability', '-> risk:attack :used:vuln -> risk:vuln'),
                    ),
                },
            },
        },

        'ou:org': {
            'default': 'identity',
            'stix': {
                'identity': {
                    'props': {
                        'name': '{-:name return($node.repr())} return(:name)',
                        'identity_class': 'return(organization)',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'sectors': '''
                            init { $list = $lib.list() }
                            -> ou:industry +:name $list.append(:name)
                            fini { if $list { return($list) } }
                        ''',
                    }
                },
                'threat-actor': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'first_seen': '+.seen $seen=.seen return($lib.stix.export.timestamp($seen.0))',
                        'last_seen': '+.seen $seen=.seen return($lib.stix.export.timestamp($seen.1))',
                        'goals': '''
                            init { $goals = $lib.list() }
                            -> ou:campaign:org -> ou:goal | uniq | +:name $goals.append(:name)
                            fini { if $goals { return($goals) } }
                        ''',
                    },
                    'rels': (
                        ('attributed-to', 'identity', ''),
                        ('located-at', 'location', ':hq -> geo:place'),
                        ('targets', 'identity', '-> ou:campaign -> risk:attack :target:org -> ou:org'),
                        ('targets', 'vulnerability', '-> ou:campaign -> risk:attack :used:vuln -> risk:vuln'),
                        # ('impersonates', 'identity', ''),
                    ),
                },
            },
        },

        'ps:person': {
            'default': 'identity',
            'stix': {
                'identity': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'identity_class': 'return(individual)',
                    }
                },
            },
        },

        'ps:contact': {
            'default': 'identity',
            'stix': {
                'identity': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'identity_class': 'return(individual)',
                    }
                },
            },
        },

        'geo:place': {
            'default': 'location',
            'stix': {
                'location': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'country': '+:loc $loc = :loc return($loc.split(".").0)',
                        'latitude': '+:latlong $latlong = :latlong return($latlong.0)',
                        'longitude': '+:latlong $latlong = :latlong return($latlong.1)',
                    },
                },
            },
        },

        'inet:ipv4': {
            'default': 'ipv4-addr',
            'stix': {
                'ipv4-addr': {
                    'props': {
                        'value': 'return($node.repr())',
                    },
                    'rels': (
                        ('belongs-to', 'autonomous-system', '-> inet:asn'),
                    ),
                }
            },
        },

        'inet:ipv6': {
            'default': 'ipv6-addr',
            'stix': {
                'ipv6-addr': {
                    'props': {
                        'value': 'return($node.repr())',
                    },
                    'rels': (
                        ('belongs-to', 'autonomous-system', '-> inet:asn'),
                    ),
                }
            },
        },

        'inet:fqdn': {
            'default': 'domain-name',
            'stix': {
                'domain-name': {
                    'props': {
                        'value': 'return($node.repr())',
                        'resolves_to_refs': '''
                            init { $refs = $lib.list() }
                            { -> inet:dns:a -> inet:ipv4 $refs.append($bundle.add($node)) }
                            { -> inet:dns:aaaa -> inet:ipv6 $refs.append($bundle.add($node)) }
                            { -> inet:dns:cname:fqdn :cname -> inet:fqdn $refs.append($bundle.add($node)) }
                            fini { if $refs { return($refs)} }
                         ''',
                    },
                }
            },
        },

        'inet:asn': {
            'default': 'autonomous-system',
            'stix': {
                'autonomous-system': {
                    'props': {
                        'name': '+:name return(:name)',
                        'number': 'return($node.value())',
                    },
                }
            },
        },

        'inet:email': {
            'default': 'email-addr',
            'stix': {
                'email-addr': {
                    'props': {
                        'value': 'return($node.repr())',
                        'display_name': '-> ps:contact +:name return(:name)',
                        'belongs_to_ref': '-> inet:web:acct return($bundle.add($node))',
                    },
                }
            },
        },

        'inet:web:acct': {
            'default': 'user-account',
            'stix': {
                'user-account': {
                    'props': {
                        'user_id': 'return(:user)',
                        'account_login': 'return(:user)',
                        'account_type': '''
                            {+:site=twitter.com return(twitter)}
                            {+:site=facebook.com return(facebook)}
                        ''',
                        'credential': '+:passwd return(:passwd)',
                        'display_name': '+:realname return(:realname)',
                        'account_created': '+:signup return($lib.stix.export.timestamp(:signup))',
                        'account_last_login': '+.seen $ival = .seen return($lib.stix.export.timestamp($ival.0))',
                        'account_first_login': '+.seen $ival = .seen return($lib.stix.export.timestamp($ival.1))',
                    },
                }
            },
        },

        'file:bytes': {
            'default': 'file',
            'stix': {
                'file': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'size': '+:size return(:size)',
                        'hashes': '''
                            init { $dict = $lib.dict() }
                            { +:md5 $dict.MD5 = :md5 }
                            { +:sha1 $dict."SHA-1" = :sha1 }
                            { +:sha256 $dict."SHA-256" = :sha256 }
                            { +:sha512 $dict."SHA-512" = :sha512 }
                            fini { if $dict { return($dict) } }
                        ''',
                        'mime_type': '+:mime return(:mime)',
                        'contains_refs': '''
                            init { $refs = $lib.list() }
                            -(refs)> *
                            $stixid = $bundle.add($node)
                            if $stixid { $refs.append($stixid) }
                            fini { if $refs { return($refs) } }
                        ''',
                    },
                },
            },
        },

        'inet:email:message': {
            'default': 'email-message',
            'stix': {
                'email-message': {
                    'props': {
                        'date': '+:date return($lib.stix.export.timestamp(:date))',
                        'subject': '+:subject return(:subject)',
                        'message_id': '-> inet:email:header +:name=message_id return(:value)',
                        'is_multipart': 'return($lib.false)',
                        'from_ref': ':from -> inet:email return($bundle.add($node))',
                        'to_refs': '''
                            init { $refs = $lib.list() }
                            { :to -> inet:email $refs.append($bundle.add($node)) }
                            fini { if $refs { return($refs) } }
                        ''',
                    },
                },
            },
        },

        'inet:url': {
            'default': 'url',
            'stix': {
                'url': {
                    'props': {
                        'value': 'return($node.repr())',
                    },
                }
            },
        },

        'syn:tag': {
            'default': 'malware',
            'stix': {
                'malware': {
                    'props': {
                        'name': '{+:title return(:title)} return($node.repr())',
                        'is_family': 'return($lib.true)',
                        'first_seen': '+.seen $seen=.seen return($lib.stix.export.timestamp($seen.0))',
                        'last_seen': '+.seen $seen=.seen return($lib.stix.export.timestamp($seen.1))',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'sample_refs': '''
                            init { $refs = $lib.list() }
                            -> file:bytes $refs.append($bundle.add($node))
                            fini { if $refs { return($refs) } }
                        ''',
                    },
                    'rels': (
                        # many of these depend on tag conventions (left for user to decide)

                        # ('controls', 'malware', '',
                        # ('authored-by', 'threat-actor', '',

                        # ('drops', 'file', '',
                        # ('drops', 'tool', '',
                        # ('drops', 'malware', '',

                        # ('downloads', 'file', '',
                        # ('downloads', 'tool', '',
                        # ('downloads', 'malware', '',

                        # ('uses', 'tool', '',
                        # ('uses', 'malware', '',
                        # ('targets', 'identity', '',
                        # ('targets', 'location', '',

                        # ('communicates-with', 'ipv4-addr',
                        # ('communicates-with', 'ipv6-addr',
                        # ('communicates-with', 'domain-name',
                        # ('communicates-with', 'url',
                    ),
                },
            },
        },

        'it:prod:soft': {
            'default': 'tool',
            'stix': {
                'tool': {
                    'props': {
                        'name': '{+:name return(:name)} return($node.repr())',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                    },
                },
            },
        },

        'it:prod:softver': {
            'default': 'tool',
            'stix': {
                'tool': {
                    'props': {
                        'name': '-> it:prod:soft {+:name return(:name)} return($node.repr())',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'tool_version': '+:vers return(:vers)',
                    },
                    'rels': (
                        # TODO
                        # ('has', 'vulnerability', ''),

                        #  depends on tag conventions
                        # ('delivers', 'malware', ''),
                        # ('drops', 'malware', ''),
                        # ('targets', 'identity', '')
                        # ('targets', 'infrastructure', '')
                        # ('targets', 'location', '')
                        # ('targets', 'vulnerability', '')
                        # ('uses', 'infrastructure', ''),
                    ),
                },
            },
        },

        'risk:vuln': {
            'default': 'vulnerability',
            'stix': {
                'vulnerability': {
                    'props': {
                        'name': '{+:name return (:name)} return ($node.repr())',
                        'description': 'if (:desc) { return (:desc) }',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'external_references': 'if :cve { $cve=:cve $cve=$cve.upper() $list=$lib.list($lib.dict(source_name=cve, external_id=$cve)) return($list) }'
                    },
                    'rels': (

                    ),
                }
            }
        },

        'media:news': {
            'default': 'report',
            'stix': {
                'report': {
                    'props': {
                        'name': 'return(:title)',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'published': 'return($lib.stix.export.timestamp(:published))',
                        'object_refs': '''
                            init { $refs = $lib.list() }
                            -(refs)> *
                            $stixid = $bundle.add($node)
                            if $stixid { $refs.append($stixid) }
                            fini { return($refs) }
                        ''',
                    },
                },
            },
        },

        'it:app:yara:rule': {
            'default': 'indicator',
            'stix': {
                'indicator': {
                    'props': {
                        'name': '+:name return(:name)',
                        'description': 'return("a yara rule")',
                        'pattern': '+:text return(:text)',
                        'pattern_type': 'return(yara)',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'valid_from': 'return($lib.stix.export.timestamp(.created))',
                    },
                },
            },
        },

        'it:app:snort:rule': {
            'default': 'indicator',
            'stix': {
                'indicator': {
                    'props': {
                        'name': '+:name return(:name)',
                        'description': 'return("a snort rule")',
                        'pattern': '+:text return(:text)',
                        'pattern_type': 'return(snort)',
                        'created': 'return($lib.stix.export.timestamp(.created))',
                        'modified': 'return($lib.stix.export.timestamp(.created))',
                        'valid_from': 'return($lib.stix.export.timestamp(.created))',
                    },
                },
            },
        },
    },
}

def _validateConfig(core, config):

    maxsize = config.get('maxsize', 10000)

    if not isinstance(maxsize, int):
        mesg = f'STIX Bundle config maxsize option must be an integer.'
        raise s_exc.BadConfValu(mesg=mesg)

    if maxsize > 10000:
        mesg = f'STIX Bundle config maxsize option must be <= 10000.'
        raise s_exc.BadConfValu(mesg=mesg)

    formmaps = config.get('forms')
    if formmaps is None:
        mesg = f'STIX Bundle config is missing "forms" mappings.'
        raise s_exc.NeedConfValu(mesg=mesg)

    for formname, formconf in formmaps.items():
        form = core.model.form(formname)
        if form is None:
            mesg = f'STIX Bundle config contains invalid form name {formname}.'
            raise s_exc.NoSuchForm(mesg=mesg)

        stixdef = formconf.get('default')
        if stixdef is None:
            mesg = f'STIX Bundle config is missing default mapping for form {formname}.'
            raise s_exc.NeedConfValu(mesg=mesg)

        if stixdef not in stix_all:
            mesg = f'STIX Bundle default mapping ({stixdef}) for {formname} is not a STIX type.'
            raise s_exc.BadConfValu(mesg=mesg)

        stixmaps = formconf.get('stix')
        if stixmaps is None:
            mesg = f'STIX Bundle config is missing STIX maps for form {formname}.'
            raise s_exc.NeedConfValu(mesg=mesg)

        if stixmaps.get(stixdef) is None:
            mesg = f'STIX Bundle config is missing STIX map for form {formname} default value {stixdef}.'
            raise s_exc.BadConfValu(mesg=mesg)

        for stixtype, stixinfo in stixmaps.items():

            if stixtype not in stix_all:
                mesg = f'STIX Bundle config has unknown STIX type {stixtype} for form {formname}.'
                raise s_exc.BadConfValu(mesg=mesg)

            stixprops = stixinfo.get('props')
            if stixprops is not None:
                for stixprop, stormtext in stixprops.items():
                    if not isinstance(stormtext, str):
                        mesg = f'STIX Bundle config has invalid prop entry {formname} {stixtype} {stixprop}.'
                        raise s_exc.BadConfValu(mesg=mesg)

            stixrels = stixinfo.get('rels')
            if stixrels is not None:
                for stixrel in stixrels:
                    if len(stixrel) != 3:
                        mesg = f'STIX Bundle config has invalid rel entry {formname} {stixtype} {stixrel}.'
                        raise s_exc.BadConfValu(mesg=mesg)

def validateStix(bundle, version='2.1'):
    ret = {
        'ok': False,
        'mesg': '',
        'result': {},
    }
    bundle = json.loads(json.dumps(bundle))
    opts = stix2validator.ValidationOptions(strict=True, version=version)
    try:
        results = stix2validator.validate_parsed_json(bundle, options=opts)
    except stix2validator.ValidationError as e:
        logger.exception('Error validating STIX bundle.')
        ret['mesg'] = f'Error validating bundle: {e}'
    else:
        ret['result'] = results.as_dict()
        ret['ok'] = bool(results.is_valid)
    return ret

@s_stormtypes.registry.registerLib
class LibStix(s_stormtypes.Lib):
    '''
    A Storm Library for interacting with Stix Version 2.1 CS02.
    '''
    _storm_locals = (  # type: ignore
        {
            'name': 'validate', 'desc': '''
            Validate a STIX Bundle.

            Notes:
                This returns a dictionary containing the following values::

                    {
                        'ok': <boolean> - False if bundle is invalid, True otherwise.
                        'mesg': <str> - An error message if there was an error when validating the bundle.
                        'results': The results of validating the bundle.
                    }
            ''',
            'type': {
                'type': 'function', '_funcname': 'validateBundle',
                'args': (
                    {'type': 'dict', 'name': 'bundle', 'desc': 'The stix bundle to validate.'},
                ),
                'returns': {'type': 'dict', 'desc': 'Results dictionary.'}
            }
        },
        {
        'name': 'lift', 'desc': '''
        Lift nodes from a STIX Bundle made by Synapse.

        Notes:
            This lifts nodes using the Node definitions embedded into the bundle when created by Synapse using
            custom extension properties.

        Examples:
            Lifting nodes from a STIX bundle::

                yield $lib.stix($bundle)
        ''',
            'type': {
                'type': 'function', '_funcname': 'liftBundle',
                'args': (
                    {'type': 'dict', 'name': 'bundle', 'desc': 'The stix bundle to lift nodes from.'},
                ),
                'returns': {'name': 'Yields', 'type': 'storm:node', 'desc': 'Yields nodes'}
            }
        },
    )
    _storm_lib_path = ('stix', )

    def __init__(self, runt, name=()):
        s_stormtypes.Lib.__init__(self, runt, name)

    def getObjLocals(self):
        return {
            'lift': self.liftBundle,
            'validate': self.validateBundle,
        }

    async def validateBundle(self, bundle):
        bundle = await s_stormtypes.toprim(bundle)
        return await s_coro.spawn((validateStix, (bundle,), {}))

    async def liftBundle(self, bundle):
        bundle = await s_stormtypes.toprim(bundle)
        for obj in bundle.get('objects', ()):
            exts = obj.get('extensions')
            if not exts:
                continue
            synx = exts.get(SYN_STIX_EXTENSION_ID)
            if not synx:  # pragma: no cover
                continue
            ndef = synx.get('synapse_ndef')
            if not ndef:  # pragma: no cover
                continue
            node = await self.runt.snap.getNodeByNdef(ndef)
            if node:
                yield node

@s_stormtypes.registry.registerLib
class LibStixExport(s_stormtypes.Lib):
    '''
    A Storm Library for exporting to STIX version 2.1 CS02.
    '''
    _storm_locals = (  # type: ignore
        {
            'name': 'bundle',
            'desc': '''
                Return a new empty STIX bundle.

                The config argument maps synapse forms to stix types and allows you to specify
                how to resolve STIX properties and relationships.  The config expects to following format::

                    {
                        "maxsize": 10000,

                        "forms": {
                            <formname>: {
                                "default": <stixtype0>,
                                "stix": {
                                    <stixtype0>: {
                                        "props": {
                                            <stix_prop_name>: <storm_with_return>,
                                            ...
                                        },
                                        "rels": (
                                            ( <relname>, <target_stixtype>, <storm> ),
                                            ...
                                        )
                                    },
                                    <stixtype1>: ...
                                },
                            },
                        },
                    },

                For example, the default config includes the following entry to map ou:campaign nodes to stix campaigns::

                    { "forms": {
                        "ou:campaign": {
                            "default": "campaign",
                            "stix": {
                                "campaign": {
                                    "props": {
                                        "name": "{+:name return(:name)} return($node.repr())",
                                        "description": "+:desc return(:desc)",
                                        "objective": "+:goal :goal -> ou:goal +:name return(:name)",
                                        "created": "return($lib.stix.export.timestamp(.created))",
                                        "modified": "return($lib.stix.export.timestamp(.created))",
                                    },
                                    "rels": (
                                        ("attributed-to", "threat-actor", ":org -> ou:org"),
                                        ("originates-from", "location", ":org -> ou:org :hq -> geo:place"),
                                        ("targets", "identity", "-> risk:attack :target:org -> ou:org"),
                                        ("targets", "identity", "-> risk:attack :target:person -> ps:person"),
                                    ),
                                },
                            },
                    }},

                Note:
                    The default config is an evolving set of mappings.  If you need to guarantee stable output please
                    specify a config.
            ''',
            'type': {
                'type': 'function', '_funcname': 'bundle',
                'args': (
                    {'type': 'dict', 'name': 'config', 'default': None, 'desc': 'The STIX bundle export config to use.'},
                ),
                'returns': {'type': 'storm:stix:bundle', 'desc': 'A new ``storm:stix:bundle`` instance.'},
            },
        },

        {
            'name': 'timestamp', 'desc': 'Format an epoch milliseconds timestamp for use in STIX output.',
             'type': {
                'type': 'function', '_funcname': 'timestamp',
                'args': (
                    {'type': 'time', 'name': 'tick', 'desc': 'The epoch milliseconds timestamp.'},
                ),
                'returns': {'type': 'str', 'desc': 'A STIX formatted timestamp string.'},
            }
        },

        {
            'name': 'config',
            'desc': 'Construct a default STIX bundle export config.',
            'type': {
                'type': 'function', '_funcname': 'config',
                'returns': {'type': 'dict', 'desc': 'A default STIX bundle export config.'},
            },
        },
    )
    _storm_lib_path = ('stix', 'export')

    def __init__(self, runt, name=()):
        s_stormtypes.Lib.__init__(self, runt, name)

    def getObjLocals(self):
        return {
            'bundle': self.bundle,
            'config': self.config,
            'timestamp': self.timestamp,
        }

    async def config(self):
        # make a new mutable config
        return json.loads(json.dumps(_DefaultConfig))

    async def bundle(self, config=None):

        if config is None:
            config = _DefaultConfig

        config = await s_stormtypes.toprim(config)
        _validateConfig(self.runt.snap.core, config)

        return StixBundle(self, self.runt, config)

    def timestamp(self, tick):
        dt = datetime.datetime.utcfromtimestamp(tick / 1000.0)
        return dt.isoformat(timespec='milliseconds') + 'Z'

stix_sdos = {
    'attack-pattern',
    'campaign',
    'course-of-action',
    'grouping',
    'identity',
    'incident',
    'indicator',
    'infrastructure',
    'intrusion-set',
    'location',
    'malware-analysis',
    'malware',
    'note',
    'observed-data',
    'opinion',
    'report',
    'threat-actor',
    'tool',
    'vulnerability',
}

stix_sros = {
    'sighting',
    'relationship',
}

stix_observables = {
    'artifact',
    'autonomous-system',
    'directory',
    'domain-name',
    'email-addr',
    'email-message',
    'file',
    'ipv4-addr',
    'ipv6-addr',
    'mac-addr',
    'mutex',
    'network-traffic',
    'process',
    'software',
    'url',
    'user-account',
    'windows-registry-key',
    'x509-certificate',
}

stix_all = set().union(stix_sdos, stix_sros, stix_observables)

@s_stormtypes.registry.registerType
class StixBundle(s_stormtypes.Prim):
    '''
    Implements the Storm API for creating and packing a STIX bundle for v2.1
    '''

    _storm_typename = 'storm:stix:bundle'
    _storm_locals = (

        {'name': 'add', 'desc': '''
        Make one or more STIX objects from a node, and add it to the bundle.

        Examples:
            Example Storm which would be called remotely via the ``callStorm()`` API::

                init { $bundle = $lib.stix.bundle() }
                #aka.feye.thr.apt1
                $bundle.add($node)
                fini { return($bundle) }
            ''',
         'type': {'type': 'function', '_funcname': 'add',
                  'args': (
                      {'name': 'node', 'desc': 'The node to make a STIX object from.', 'type': 'node'},
                      {'name': 'stixtype', 'type': 'str', 'default': None,
                       'desc': 'The explicit name of the STIX type to map the node to. This will override the default'
                               ' mapping.', },
                  ),
                  'returns': {'type': 'str', 'desc': 'The stable STIX id of the added object.'}}},

        {'name': 'pack', 'desc': 'Return the bundle as a STIX JSON object.',
         'type': {'type': 'function', '_funcname': 'pack',
                  'args': (),
                  'returns': {'type': 'dict', }}},

    )

    def __init__(self, libstix, runt, config, path=None):

        s_stormtypes.Prim.__init__(self, 'bundle', path=path)

        self.runt = runt
        self.libstix = libstix
        self.config = config
        self.locls.update(self.getObjLocals())
        self.objs = {}  # id -> STIX obj(dict)
        self.maxsize = config.get('maxsize', 10000)

    async def value(self):
        return self.pack()

    def getObjLocals(self):
        return {
            'add': self.add,
            'pack': self.pack,
        }

    # TODO config modification helpers
    # async def addPropMap(self, formname, stixtype, propname, stormtext):
    # async def addRelsMap(self, formname, stixtype, relname, targtype,  stormtext):

    async def add(self, node, stixtype=None):

        if len(self.objs) >= self.maxsize:
            mesg = f'STIX Bundle is at maxsize ({self.maxsize}).'
            raise s_exc.StormRuntimeError(mesg=mesg)

        if not isinstance(node, s_node.Node):
            await self.runt.warnonce('STIX bundle add() method requires a node.')
            return None

        formconf = self.config['forms'].get(node.form.name)
        if formconf is None:
            await self.runt.warnonce(f'STIX bundle has no config for mapping {node.form.name}.')
            return None

        if stixtype is None:
            stixtype = formconf.get('default')

        # cyber observables have UUIDv5 the rest have UUIDv4
        if stixtype in stix_observables:
            stixid = f'{stixtype}--{uuid5(node.ndef)}'
        else:
            stixid = f'{stixtype}--{uuid4(node.ndef)}'

        if self.objs.get(stixid) is not None:
            return stixid

        stixconf = formconf['stix'].get(stixtype)
        if stixconf is None:
            await self.runt.warnonce(f'STIX bundle config has no config to map {node.form.name} to {stixtype}.')
            return None

        stixitem = self.objs.get(stixid)
        if stixitem is None:
            stixitem = self.objs[stixid] = self._initStixItem(stixid, stixtype, node)

        props = stixconf.get('props')
        if props is not None:

            for name, storm in props.items():
                valu = await self._callStorm(storm, node)
                if valu is s_common.novalu:
                    continue

                stixitem[name] = valu

        for (relname, reltype, relstorm) in stixconf.get('rels', ()):
            async for relnode, relpath in node.storm(self.runt, relstorm):
                n2id = await self.add(relnode, stixtype=reltype)
                await self._addRel(stixid, relname, n2id)

        return stixid

    def _initStixItem(self, stixid, stixtype, node):
        ndef = json.loads(json.dumps(node.ndef))
        retn = {
            'id': stixid,
            'type': stixtype,
            'spec_version': '2.1',
            'extensions': {
                SYN_STIX_EXTENSION_ID: {
                    "extension_type": "property-extension",
                    'synapse_ndef': ndef
                }
            }

        }
        return retn

    async def _addRel(self, srcid, reltype, targid):

        stixid = f'relationship--{uuid4((srcid, reltype, targid))}'
        tstamp = self.libstix.timestamp(s_common.now())

        obj = {
            'id': stixid,
            'type': 'relationship',
            'relationship_type': reltype,
            'created': tstamp,
            'modified': tstamp,
            'source_ref': srcid,
            'target_ref': targid,
            'spec_version': '2.1',
        }

        self.objs[stixid] = obj
        return stixid

    def _getSynapseExtensionDefinition(self):
        ret = {
            'id': SYN_STIX_EXTENSION_ID,
            'type': 'extension-definition',
            'spec_version': '2.1',
            'name': 'Vertex Project Synapse',
            'description': 'Synapse specific STIX 2.1 extensions.',
            'created': '2021-04-29T13:40:00.000Z',
            'modified': '2021-04-29T13:40:00.000Z',
            'schema': 'The Synapse Extensions for Stix 2.1',
            'version': '1.0',
            'extension_types': [
                'property-extension',
            ],
        }
        return ret

    def pack(self):
        objects = list(self.objs.values())
        objects.insert(0, self._getSynapseExtensionDefinition())
        bundle = {
            'type': 'bundle',
            'id': f'bundle--{uuid4()}',
            'objects': objects
        }
        return bundle

    async def _callStorm(self, text, node):

        opts = {'vars': {'bundle': self}}
        query = await self.runt.snap.core.getStormQuery(text)
        async with self.runt.getSubRuntime(query, opts=opts) as runt:

            async def genr():
                yield node, runt.initPath(node)

            try:
                async for _ in runt.execute(genr=genr()):
                    await asyncio.sleep(0)
            except s_stormctrl.StormReturn as e:
                return await s_stormtypes.toprim(e.item)

        return s_common.novalu




import sys
import argparse


import synapse.exc as s_exc
import synapse.lib.output as s_output
import synapse.lib.certdir as s_certdir

descr = '''
Command line tool to generate simple x509 certs
'''

def main(argv, outp=None):

    if outp is None:  # pragma: no cover
        outp = s_output.OutPut()

    pars = argparse.ArgumentParser(prog='easycert', description=descr)
    pars.add_argument('--certdir', default='~/.syn/certs', help='Directory for certs/keys')
    pars.add_argument('--importfile', choices=('cas', 'hosts', 'users'), help='import certs and/or keys into local certdir')

    pars.add_argument('--ca', default=False, action='store_true', help='mark the certificate as a CA/CRL signer')
    pars.add_argument('--p12', default=False, action='store_true', help='mark the certificate as a p12 archive')
    pars.add_argument('--server', default=False, action='store_true', help='mark the certificate as a server')
    pars.add_argument('--server-sans', help='server cert subject alternate names')

    pars.add_argument('--csr', default=False, action='store_true', help='generate a cert signing request')
    pars.add_argument('--sign-csr', default=False, action='store_true', help='sign a cert signing request')
    pars.add_argument('--signas', help='sign the new cert with the given cert name')

    pars.add_argument('name', help='common name for the certificate (or filename for CSR signing)')

    opts = pars.parse_args(argv)

    cdir = s_certdir.CertDir(path=opts.certdir)

    try:

        if opts.importfile:
            cdir.importFile(opts.name, opts.importfile, outp=outp)
            return 0

        if opts.p12:

            cdir.genClientCert(opts.name, outp=outp)
            return 0

        if opts.sign_csr:

            if opts.signas is None:
                outp.printf('--sign-csr requires --signas')
                return -1

            xcsr = cdir._loadCsrPath(opts.name)
            if xcsr is None:
                outp.printf('csr not found: %s' % (opts.name,))
                return -1

            if opts.server:
                cdir.signHostCsr(xcsr, opts.signas, outp=outp)
                return 0

            cdir.signUserCsr(xcsr, opts.signas, outp=outp)
            return 0

        if opts.csr:

            if opts.ca:
                raise NotImplementedError('CSR for CA cert not supported (yet)')

            if opts.server:
                cdir.genHostCsr(opts.name, outp=outp)
                return 0

            cdir.genUserCsr(opts.name, outp=outp)
            return 0

        if opts.ca:
            cdir.genCaCert(opts.name, signas=opts.signas, outp=outp)
            return 0

        if opts.server:
            cdir.genHostCert(opts.name, signas=opts.signas, outp=outp, sans=opts.server_sans)
            return 0

        cdir.genUserCert(opts.name, signas=opts.signas, outp=outp)
        return 0

    except s_exc.DupFileName as e:
        outp.printf('file exists: %s' % (e.errinfo.get('path'),))
        return -1

if __name__ == '__main__':  # pragma: no cover
    sys.exit(main(sys.argv[1:]))



import synapse.common as s_common
import synapse.lib.output as s_output

def main(argv, outp=None):
    if outp is None:  # pragma: no cover
        outp = s_output.OutPut()

    outp.printf(s_common.guid())

if __name__ == '__main__':  # pragma: no cover
    import sys
    sys.exit(main(sys.argv))



import os
import sys
import asyncio

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.output as s_output
import synapse.lib.dyndeps as s_dyndeps

def getStemCell(dirn):

    if not os.path.isdir(dirn):
        mesg = f'Directory {dirn} does not exist!'
        raise s_exc.NoSuchDir(mesg=mesg)

    ctorname = os.getenv('SYN_STEM_CELL_CTOR')

    cellyaml = os.path.join(dirn, 'cell.yaml')

    if os.path.isfile(cellyaml):
        conf = s_common.yamlload(cellyaml)
        ctorname = conf.get('cell:ctor', ctorname)

    if ctorname is not None:
        ctorname = ctorname.strip()
        ctor = s_dyndeps.getDynLocal(ctorname)
        if ctor is None:
            raise s_exc.NoSuchCtor(mesg=f'Unable to resolve ctor [{ctorname}]', ctor=ctorname)
        return ctor

    mesg = f'No such file: {cellyaml} and SYN_STEM_CELL_CTOR environmt variable is not set.'
    raise s_exc.NoSuchFile(mesg=mesg, path=cellyaml)

async def main(argv, outp=s_output.stdout):  # pragma: no cover
    ctor = getStemCell(argv[0])
    return await ctor.execmain(argv, outp=outp)

if __name__ == '__main__':  # pragma: no cover
    sys.exit(asyncio.run(main(sys.argv[1:])))



import time

class RateLimit:
    '''
    A RateLimit class may be used to detect/enforce rate limits.

    Example:

        # allow 20 uses per 10 sec ( 2/sec )
        rlimit = RateLimit(20,10)

    Notes:

        It is best ( even in a "calls per day" type config ) to
        specify a smaller "per" to force rate "smoothing".

    '''
    def __init__(self, rate, per):
        self.rate = float(rate)
        self.per = float(per)

        self.lasttick = time.time()
        self.allowance = float(rate)

        self.persec = float(rate) / float(per)
        self.onetick = float(per) / float(rate)

    def allows(self):
        '''
        Returns True if the rate limit has not been reached.

        Example:

            if not rlimit.allows():
                rasie RateExceeded()

            # ok to go...

        '''
        tick = time.time()
        passed = tick - self.lasttick

        self.allowance = min(self.rate, self.allowance + (passed * self.persec))

        self.lasttick = tick

        if self.allowance < 1.0:
            return False

        self.allowance -= 1.0
        return True



import os

import unittest.mock as mock

import synapse.common as s_common
import synapse.tests.utils as s_test

import synapse.lib.cell as s_cell
import synapse.lib.hive as s_hive
import synapse.lib.msgpack as s_msgpack

import synapse.tools.hive.save as s_hivesave

class HiveSaveTest(s_test.SynTest):

    async def test_tools_hivesave_vercheck(self):
        with self.getTestDir() as dirn:

            hivepath0 = os.path.join(dirn, 'hivesave0.mpk')

            async with self.getTestHiveDmon() as dmon:
                hurl = self.getTestUrl(dmon, 'hive')

                argv = [hurl, hivepath0]

                def _getOldSynVers(self):
                    return (0, 0, 0)

                with mock.patch('synapse.telepath.Proxy._getSynVers', _getOldSynVers):
                    outp = self.getTestOutp()
                    retn = await s_hivesave.main(argv, outp=outp)
                    outp.expect('Hive version 0.0.0 is outside of the hive.save supported range')
                    self.eq(1, retn)

    async def test_tools_hivesave(self):

        with self.getTestDir() as dirn:

            hivepath0 = os.path.join(dirn, 'hivesave0.mpk')
            yamlpath0 = os.path.join(dirn, 'hivesave0.yaml')

            async with self.getTestHiveDmon() as dmon:

                hive = dmon.shared.get('hive')
                await hive.set(('baz', 'faz'), 'visi')

                hurl = self.getTestUrl(dmon, 'hive')

                retn = await s_hivesave.main([hurl, hivepath0])
                self.eq(0, retn)

                tree = s_msgpack.loadfile(hivepath0)
                self.eq('visi', tree['kids']['baz']['kids']['faz']['value'])

                await s_hivesave.main(['--path', 'baz', '--yaml', hurl, yamlpath0])

                tree = s_common.yamlload(yamlpath0)
                self.eq('visi', tree['kids']['faz']['value'])

            hivepath1 = os.path.join(dirn, 'hivesave1.mpk')
            yamlpath1 = os.path.join(dirn, 'hivesave1.yaml')

            path = os.path.join(dirn, 'cell')
            async with await s_cell.Cell.anit(path) as cell:

                await cell.hive.set(('hehe', 'haha'), 20)
                curl = cell.getLocalUrl()

                await s_hivesave.main([curl, hivepath1])
                tree = s_msgpack.loadfile(hivepath1)
                self.eq(20, tree['kids']['hehe']['kids']['haha']['value'])

                await s_hivesave.main(['--path', 'hehe', '--yaml', curl, yamlpath1])
                tree = s_common.yamlload(yamlpath1)

                self.eq(20, tree['kids']['haha']['value'])



import synapse.exc as s_exc
import synapse.tests.utils as s_test

xml0 = '''
<?xml version="1.0"?>
<data>
    <country name="Liechtenstein">
        <rank>1</rank>
        <year>2008</year>
        <gdppc>141100</gdppc>
        <neighbor name="Austria" direction="E"/>
        <neighbor name="Switzerland" direction="W"/>
    </country>
    <country name="Singapore">
        <rank>4</rank>
        <year>2011</year>
        <gdppc>59900</gdppc>
        <neighbor name="Malaysia" direction="N"/>
    </country>
    <country name="Panama">
        <rank>68</rank>
        <year>2011</year>
        <gdppc>13600</gdppc>
        <neighbor name="Costa Rica" direction="W"/>
        <neighbor name="Colombia" direction="E"/>
    </country>
</data>
'''.strip()

class XmlTest(s_test.SynTest):

    async def test_stormlib_xml(self):

        async with self.getTestCore() as core:
            valu = await core.callStorm('''
                $retn = $lib.list()
                $root = $lib.xml.parse($xmltext)
                for $elem in $root {
                    $retn.append((
                        $elem.name,
                        $elem.attrs.name,
                        $elem.get(rank).text,
                    ))
                }
                return($retn)
            ''', opts={'vars': {'xmltext': xml0}})
            self.eq(valu, (
                ('country', 'Liechtenstein', '1'),
                ('country', 'Singapore', '4'),
                ('country', 'Panama', '68'),
            ))

            valu = await core.callStorm('''
                $retn = $lib.list()
                $root = $lib.xml.parse($xmltext)
                for $elem in $root.find(country) {
                    $retn.append((
                        $elem.name,
                        $elem.attrs.name,
                        $elem.get(rank).text,
                    ))
                }
                return($retn)
            ''', opts={'vars': {'xmltext': xml0}})
            self.eq(valu, (
                ('country', 'Liechtenstein', '1'),
                ('country', 'Singapore', '4'),
                ('country', 'Panama', '68'),
            ))

            valu = await core.callStorm('''
                $retn = $lib.list()
                $root = $lib.xml.parse($xmltext)
                for $elem in $root.find(rank) {
                    $retn.append($elem.text)
                }
                return($retn)
            ''', opts={'vars': {'xmltext': xml0}})
            self.eq(valu, ('1', '4', '68'))

            valu = await core.callStorm('''
                $retn = $lib.list()
                $root = $lib.xml.parse($xmltext)
                for $elem in $root.find(rank, nested=$lib.false) {
                    $retn.append($elem.text)
                }
                return($retn)
            ''', opts={'vars': {'xmltext': xml0}})
            self.eq(valu, ())

            valu = await core.callStorm('''
                $retn = $lib.list()
                $root = $lib.xml.parse($xmltext)
                for $elem in $root.find(rank, nested=$lib.false) {
                    $retn.append($elem.text)
                }
                return($retn)
            ''', opts={'vars': {'xmltext': xml0}})
            self.eq(valu, [])

            with self.raises(s_exc.BadArg):
                opts = {'vars': {'bytes': b'asdf'}}
                await core.callStorm('$lib.xml.parse($bytes)', opts=opts)



import gc
import os
import atexit
import signal
import asyncio
import inspect
import logging
import weakref
import contextlib
import collections

if __debug__:
    import traceback

import synapse.exc as s_exc
import synapse.glob as s_glob

import synapse.lib.coro as s_coro

logger = logging.getLogger(__name__)

OMIT_FINI_WARNS = os.environ.get('SYNDEV_OMIT_FINI_WARNS', False)

def _fini_atexit():  # pragma: no cover

    for item in gc.get_objects():

        if not isinstance(item, Base):
            continue

        if not item.anitted:
            continue

        if item.isfini:
            continue

        if not item._fini_atexit and not OMIT_FINI_WARNS:
            if __debug__:
                logger.debug(f'At exit: Missing fini for {item}')
                for depth, call in enumerate(item.call_stack[:-2]):
                    logger.debug(f'{depth+1:3}: {call.strip()}')
            continue

        try:
            if __debug__:
                logger.debug('At exit: Calling fini for %r', item)
            rv = item.fini()
            if s_coro.iscoro(rv):
                # Try to run the fini on its loop
                loop = item.loop
                if not loop.is_running():
                    continue
                loop.create_task(rv)

        except Exception:
            logger.exception('atexit fini fail: %r' % (item,))

atexit.register(_fini_atexit)

class Base:
    '''
    Base class for Synapse objects.

    Acts as an observable, enables async init and fini.

    Example:

        class Foo(Base):

            async def __anit__(self, x, y):

                await Base.__anit__(self)

                await stuff(x, y)

        foo = await Foo.anit(10)

    Note:
        One should not create instances directly via its initializer, i.e. Base().  One shall always use the class
        method anit.
    '''
    def __init__(self):
        self.anitted = False
        assert inspect.stack()[1].function == 'anit', 'Objects from Base must be constructed solely via "anit"'

    @classmethod
    async def anit(cls, *args, **kwargs):

        # sneak in a quick loop check here for convenience
        if s_glob._glob_loop is None:
            s_glob.initloop()

        self = cls()

        try:

            await self.__anit__(*args, **kwargs)

        except (asyncio.CancelledError, Exception):
            if self.anitted:
                await self.fini()

            raise

        try:
            await self.postAnit()
        except (asyncio.CancelledError, Exception):
            logger.exception('Error during postAnit callback.')
            await self.fini()
            raise

        return self

    async def __anit__(self):

        self.loop = asyncio.get_running_loop()
        if __debug__:
            import synapse.lib.threads as s_threads  # avoid import cycle
            self.tid = s_threads.iden()
            self.call_stack = traceback.format_stack()  # For cleanup debugging

        if object.__getattribute__(self, 'anitted') is True:
            # The Base has already been anitted. This allows a class to treat
            # multiple Base objects as a mixin and __anit__ themselves without
            # smashing fini or event handlers from the others.
            return

        self.isfini = False
        self.anitted = True  # For assertion purposes
        self.finievt = None
        self.entered = False

        # hold a weak ref to other bases we should fini if they
        # are still around when we go down...
        self.tofini = weakref.WeakSet()

        self._syn_funcs = collections.defaultdict(list)

        self._syn_refs = 1  # one ref for the ctor
        self._syn_links = []
        self._fini_funcs = []
        self._fini_atexit = False
        self._active_tasks = set()  # the free running tasks associated with me

    async def postAnit(self):
        '''
        Method called after self.__anit__() has completed, but before anit() returns the object to the caller.
        '''
        pass

    async def enter_context(self, item):
        '''
        Modeled on Python's contextlib.ExitStack.enter_context.  Enters a new context manager and adds its __exit__()
        and __aexit__ method to its onfini handlers.

        Returns:
            The result of item’s own __aenter__ or __enter__() method.
        '''
        async def fini():
            meth = getattr(item, '__aexit__', None)
            if meth is not None:
                return await meth(None, None, None)

            meth = getattr(item, '__exit__', None)
            if meth is not None:
                return meth(None, None, None)

        self.onfini(fini)

        entr = getattr(item, '__aenter__', None)
        if entr is not None:
            async def fini():
                await item.__aexit__(None, None, None)
            self.onfini(fini)
            return await entr()

        entr = getattr(item, '__enter__', None)
        assert entr is not None

        async def fini():
            item.__exit__(None, None, None)
        self.onfini(fini)
        return entr()

    def onfini(self, func):
        '''
        Add a function/coroutine/Base to be called on fini().
        '''
        if isinstance(func, Base):
            self.tofini.add(func)
            return

        assert self.anitted
        self._fini_funcs.append(func)

    async def __aenter__(self):
        assert asyncio.get_running_loop() == self.loop
        self.entered = True
        return self

    async def __aexit__(self, exc, cls, tb):
        # Either there should be no running loop or we shall be on the right one
        try:
            assert asyncio.get_running_loop() == self.loop
        except RuntimeError:
            pass

        await self.fini()

    def incref(self):
        '''
        Increment the reference count for this base.  This API may be optionally used to control fini().
        '''
        self._syn_refs += 1
        return self._syn_refs

    def link(self, func):
        '''
        Add a callback function to receive *all* events.

        Example:

            base1 = Base()
            base2 = Base()

            base1.link( base2.dist )

            # all events on base1 are also propagated on base2

        '''
        self._syn_links.append(func)

    def unlink(self, func):
        '''
        Remove a callback function previously added with link()

        Example:

            base.unlink( callback )

        '''
        if func in self._syn_links:
            self._syn_links.remove(func)

    def on(self, evnt, func, base=None):
        '''
        Add an base function callback for a specific event with optional filtering.  If the function returns a
        coroutine, it will be awaited.

        Args:
            evnt (str):         An event name
            func (function):    A callback function to receive event tufo

        Examples:

            Add a callback function and fire it:

                async def baz(event):
                    x = event[1].get('x')
                    y = event[1].get('y')
                    return x + y

                d.on('foo', baz)

                # this fire triggers baz...
                await d.fire('foo', x=10, y=20)

        Returns:
            None:
        '''
        funcs = self._syn_funcs[evnt]
        if func in funcs:
            return

        funcs.append(func)

        if base is not None:

            def fini():
                self.off(evnt, func)

            base.onfini(fini)

    def off(self, evnt, func):
        '''
        Remove a previously registered event handler function.

        Example:

            base.off( 'foo', onFooFunc )

        '''
        funcs = self._syn_funcs.get(evnt)
        if funcs is None:
            return

        try:
            funcs.remove(func)
        except ValueError:
            pass

    async def fire(self, evtname, **info):
        '''
        Fire the given event name on the Base.
        Returns a list of the return values of each callback.

        Example:

            for ret in d.fire('woot',foo='asdf'):
                print('got: %r' % (ret,))

        '''
        event = (evtname, info)
        if self.isfini:
            return event

        await self.dist(event)
        return event

    async def dist(self, mesg):
        '''
        Distribute an existing event tuple.

        Args:
            mesg ((str,dict)):  An event tuple.

        Example:

            await base.dist( ('foo',{'bar':'baz'}) )

        '''
        if self.isfini:
            return ()

        ret = []
        for func in self._syn_funcs.get(mesg[0], ()):

            try:
                ret.append(await s_coro.ornot(func, mesg))
            except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                raise
            except Exception:
                logger.exception('base %s error with mesg %s', self, mesg)

        for func in self._syn_links:
            try:
                ret.append(await s_coro.ornot(func, mesg))
            except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                raise
            except Exception:
                logger.exception('base %s error with mesg %s', self, mesg)

        return ret

    async def _kill_active_tasks(self):

        if not self._active_tasks:
            return

        for task in self._active_tasks.copy():

            task.cancel()
            try:
                await task
            except (asyncio.CancelledError, Exception):
                # The taskDone callback will emit the exception.  No need to repeat
                pass

    async def fini(self):
        '''
        Shut down the object and notify any onfini() coroutines.

        Returns:
            Remaining ref count
        '''
        assert self.anitted, f'{self.__class__.__name__} initialized improperly.  Must use Base.anit class method.'

        if self.isfini:
            return

        if __debug__:
            import synapse.lib.threads as s_threads  # avoid import cycle
            assert s_threads.iden() == self.tid

        self._syn_refs -= 1
        if self._syn_refs > 0:
            return self._syn_refs

        self.isfini = True

        for base in list(self.tofini):
            await base.fini()

        await self._kill_active_tasks()

        for fini in self._fini_funcs:
            try:
                await s_coro.ornot(fini)
            except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                raise
            except Exception:
                logger.exception(f'{self} - fini function failed: {fini}')

        self._syn_funcs.clear()
        self._fini_funcs.clear()

        fevt = self.finievt

        if fevt is not None:
            fevt.set()

        return 0

    @contextlib.contextmanager
    def onWith(self, evnt, func):
        '''
        A context manager which can be used to add a callback and remove it when
        using a ``with`` statement.

        Args:
            evnt (str):         An event name
            func (function):    A callback function to receive event tufo
        '''
        self.on(evnt, func)
        # Allow exceptions to propagate during the context manager
        # but ensure we cleanup our temporary callback
        try:
            yield self
        finally:
            self.off(evnt, func)

    async def waitfini(self, timeout=None):
        '''
        Wait for the base to fini()

        Returns:
            None if timed out, True if fini happened

        Example:

            base.waitfini(timeout=30)

        '''

        if self.isfini:
            return True

        if self.finievt is None:
            self.finievt = asyncio.Event()

        return await s_coro.event_wait(self.finievt, timeout)

    def schedCoro(self, coro):
        '''
        Schedules a free-running coroutine to run on this base's event loop.  Kills the coroutine if Base is fini'd.
        It does not pend on coroutine completion.

        Precondition:
            This function is *not* threadsafe and must be run on the Base's event loop

        Returns:
            asyncio.Task: An asyncio.Task object.

        '''
        import synapse.lib.provenance as s_provenance  # avoid import cycle

        if __debug__:
            assert inspect.isawaitable(coro)
            import synapse.lib.threads as s_threads  # avoid import cycle
            assert s_threads.iden() == self.tid

        task = self.loop.create_task(coro)

        # In rare cases, (Like this function being triggered from call_soon_threadsafe), there's no task context
        if asyncio.current_task():
            s_provenance.dupstack(task)

        def taskDone(task):
            self._active_tasks.remove(task)
            try:
                if not task.done():
                    task.result()
            except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                pass
            except Exception:
                logger.exception('Task %s scheduled through Base.schedCoro raised exception', task)

        self._active_tasks.add(task)
        task.add_done_callback(taskDone)

        return task

    def schedCallSafe(self, func, *args, **kwargs):
        '''
        Schedule a function to run as soon as possible on the same event loop that this Base is running on.

        This function does *not* pend on the function completion.

        Args:
            func:
            *args:
            **kwargs:

        Notes:
            This method may be called from outside of the event loop on a different thread.

        Returns:
            concurrent.futures.Future: A Future representing the eventual function execution.
        '''
        def real():
            return func(*args, **kwargs)
        return self.loop.call_soon_threadsafe(real)

    def schedCoroSafe(self, coro):
        '''
        Schedules a coroutine to run as soon as possible on the same event loop that this Base is running on.

        This function does *not* pend on coroutine completion.

        Notes:
            This method may be run outside the event loop on a different thread.

        Returns:
            concurrent.futures.Future: A Future representing the eventual coroutine execution.
        '''
        return self.loop.call_soon_threadsafe(self.schedCoro, coro)

    def schedCoroSafePend(self, coro):
        '''
        Schedules a coroutine to run as soon as possible on the same event loop that this Base is running on

        Note:
            This method may *not* be run inside an event loop
        '''
        if __debug__:
            import synapse.lib.threads as s_threads  # avoid import cycle
            assert s_threads.iden() != self.tid

        task = asyncio.run_coroutine_threadsafe(coro, self.loop)
        return task.result()

    async def addSignalHandlers(self):
        '''
        Register SIGTERM/SIGINT signal handlers with the ioloop to fini this object.
        '''

        def sigterm():
            print('Caught SIGTERM, shutting down.')
            asyncio.create_task(self.fini())

        def sigint():
            print('Caught SIGINT, shutting down.')
            asyncio.create_task(self.fini())

        loop = asyncio.get_running_loop()
        loop.add_signal_handler(signal.SIGINT, sigint)
        loop.add_signal_handler(signal.SIGTERM, sigterm)

    async def main(self):
        '''
        Helper function to setup signal handlers for this base as the main object.
        ( use base.waitfini() to block )

        NOTE: This API may only be used when the ioloop is *also* the main thread.
        '''
        await self.addSignalHandlers()
        return await self.waitfini()

    def waiter(self, count, *names):
        '''
        Construct and return a new Waiter for events on this base.

        Example:

            # wait up to 3 seconds for 10 foo:bar events...

            waiter = base.waiter(10,'foo:bar')

            # .. fire thread that will cause foo:bar events

            events = await waiter.wait(timeout=3)

            if events == None:
                # handle the timeout case...

            for event in events:
                # parse the events if you need...

        NOTE: use with caution... it's easy to accidentally construct
              race conditions with this mechanism ;)

        '''
        return Waiter(self, count, self.loop, *names)

class Waiter:
    '''
    A helper to wait for a given number of events on a Base.
    '''
    def __init__(self, base, count, *names):
        self.base = base
        self.names = names
        self.count = count
        self.event = asyncio.Event()

        self.events = []

        for name in names:
            base.on(name, self._onWaitEvent)

        if not names:
            base.link(self._onWaitEvent)

    def _onWaitEvent(self, mesg):
        self.events.append(mesg)
        if len(self.events) >= self.count:
            self.event.set()

    async def wait(self, timeout=None):
        '''
        Wait for the required number of events and return them or None on timeout.

        Example:

            evnts = waiter.wait(timeout=30)

            if evnts == None:
                handleTimedOut()
                return

            for evnt in evnts:
                doStuff(evnt)

        '''
        try:

            retn = await s_coro.event_wait(self.event, timeout)
            if not retn:
                return None

            return self.events

        finally:
            self.fini()

    def fini(self):

        for name in self.names:
            self.base.off(name, self._onWaitEvent)

        if not self.names:
            self.base.unlink(self._onWaitEvent)
        del self.event

class BaseRef(Base):
    '''
    An object for managing multiple Base instances by name.
    '''
    async def __anit__(self, ctor=None):
        await Base.__anit__(self)
        self.ctor = ctor
        self.base_by_name = {}
        self.onfini(self._onBaseRefFini)

    async def _onBaseRefFini(self):
        await asyncio.gather(*[base.fini() for base in self.vals()])

    def put(self, name, base):
        '''
        Add a Base (or sub-class) to the BaseRef by name.

        Args:
            name (str): The name/iden of the Base
            base (Base): The Base instance

        Returns:
            (None)
        '''
        async def fini():
            if self.base_by_name.get(name) is base:
                self.base_by_name.pop(name, None)

        # Remove myself from BaseRef when I fini
        base.onfini(fini)
        self.base_by_name[name] = base

    def pop(self, name):
        '''
        Remove and return a Base from the BaseRef.

        Args:
            name (str): The name/iden of the Base instance

        Returns:
            (Base): The named base ( or None )
        '''
        return self.base_by_name.pop(name, None)

    def get(self, name):
        '''
        Retrieve a Base instance by name.

        Args:
            name (str): The name/iden of the Base

        Returns:
            (Base): The Base instance (or None)
        '''
        return self.base_by_name.get(name)

    async def gen(self, name):
        '''
        Atomically get/gen a Base and incref.
        (requires ctor during BaseRef init)

        Args:
            name (str): The name/iden of the Base instance.
        '''
        if self.ctor is None:
            raise s_exc.NoSuchCtor(name=name, mesg='BaseRef.gen() requires ctor')

        base = self.base_by_name.get(name)

        if base is None:
            base = await self.ctor(name)
            self.put(name, base)
        else:
            base.incref()

        return base

    def vals(self):
        return list(self.base_by_name.values())

    def items(self):
        return list(self.base_by_name.items())

    def __iter__(self):
        # make a copy during iteration to prevent dict
        # change during iteration exceptions
        return iter(list(self.base_by_name.values()))

async def schedGenr(genr, maxsize=100):
    '''
    Schedule a generator to run on a separate task and yield results to this task (pipelined generator).
    '''
    q = asyncio.Queue(maxsize=maxsize)

    async def genrtask(base):
        try:
            async for item in genr:
                await q.put((True, item))

            await q.put((False, None))

        except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
            raise

        except Exception:
            if not base.isfini:
                await q.put((False, None))
            raise

    async with await Base.anit() as base:

        task = base.schedCoro(genrtask(base))

        while not base.isfini:

            ok, retn = await q.get()

            if ok:
                yield retn
                # since we are a pipeline, yield every time...
                await asyncio.sleep(0)
                continue

            await task
            return

async def main(coro):  # pragma: no cover
    base = await coro
    if isinstance(base, Base):
        async with base:
            await base.main()



import sys
import time
import asyncio
import threading

import synapse.exc as s_exc
import synapse.glob as s_glob
import synapse.lib.coro as s_coro
import synapse.tests.utils as s_t_utils

class FakeError(Exception): pass

def spawnfunc(x, y=10):
    return x + y

def spawnsleep(n=10):
    time.sleep(n)
    return True

def spawnfakeit():
    raise FakeError()

def spawnexit():
    sys.exit(0)

def chkpool():
    return s_coro.forkpool is None

class CoroTest(s_t_utils.SynTest):

    async def test_coro_event(self):

        evnt = s_coro.Event()

        evnt.set()
        self.true(await evnt.timewait())

        evnt.clear()
        self.false(await evnt.timewait(timeout=0.001))

    async def test_coro_iscoro(self):

        async def agen():
            yield 42

        def genr():
            yield 'woot'

        async def woot():
            return 10

        item = woot()
        self.true(s_coro.iscoro(item))

        await item

        self.false(s_coro.iscoro(genr()))
        self.false(s_coro.iscoro(agen()))

    async def test_coro_genrhelp(self):

        @s_coro.genrhelp
        async def woot():
            yield 1
            yield 2
            yield 3

        self.none(await woot().spin())
        self.eq([1, 2, 3], await woot().list())

    async def test_executor(self):

        def func(*args, **kwargs):
            tid = threading.get_ident()
            return tid, args, kwargs

        future = s_coro.executor(func, 1, key='valu')
        tid, args, kwargs = await future
        # Ensure that we were not executed on the ioloop thread
        self.ne(tid, s_glob._glob_thrd.ident)
        # Ensure that args are passed as expected
        self.eq(args, (1,))
        self.eq(kwargs, {'key': 'valu'})

        async def afunc():
            tid = threading.get_ident()
            return tid
        # Ensure a generic coroutine is executed on the ioloop thread
        self.eq(s_glob._glob_thrd.ident, await afunc())

    async def test_lib_coro_spawn(self):

        todo = (spawnfunc, (20,), {'y': 30})
        self.eq(50, await s_coro.spawn(todo))

        todo = (spawnsleep, (1.5,), {})
        self.true(await s_coro.spawn(todo))

        todo = (spawnsleep, (), {})
        with self.raises(asyncio.TimeoutError):
            await s_coro.spawn(todo, timeout=0.1)

        todo = (spawnfakeit, (), {})
        with self.raises(FakeError):
            await s_coro.spawn(todo)

        todo = (spawnexit, (), {})
        with self.raises(s_exc.SpawnExit):
            await s_coro.spawn(todo)

    async def test_lib_coro_forked(self):

        self.true(await s_coro.forked(chkpool))

        self.eq(50, await s_coro.forked(spawnfunc, 20, y=30))

        with self.raises(FakeError):
            await s_coro.forked(spawnfakeit)

        self.eq(50, await s_coro.forked(spawnfunc, 20, y=30))

        def newp():
            return 23

        with self.raises(Exception):
            await s_coro.forked(newp)



import os
import signal
import asyncio
import logging
import threading
import faulthandler

logger = logging.getLogger(__name__)

_glob_loop = None
_glob_thrd = None


def _asynciostacks(*args, **kwargs):  # pragma: no cover
    '''
    A signal handler used to print asyncio task stacks and thread stacks.
    '''
    print(80 * '*')
    print('Asyncio tasks stacks:')
    tasks = asyncio.all_tasks(_glob_loop)
    for task in tasks:
        task.print_stack()
    print(80 * '*')
    print('Faulthandler stack frames per thread:')
    faulthandler.dump_traceback()
    print(80 * '*')

def _threadstacks(*args, **kwargs):  # pragma: no cover
    '''
    A signal handler used to print thread stacks.
    '''
    print(80 * '*')
    print('Faulthandler stack frames per thread:')
    faulthandler.dump_traceback()
    print(80 * '*')

signal.signal(signal.SIGUSR1, _threadstacks)
signal.signal(signal.SIGUSR2, _asynciostacks)

def initloop():

    global _glob_loop
    global _glob_thrd

    # if there's no global loop....
    if _glob_loop is None:

        # check if it's us....
        try:
            _glob_loop = asyncio.get_running_loop()
            # if we get here, it's us!
            _glob_thrd = threading.currentThread()
            # Enable debug and greedy coro collection
            setGreedCoro(_glob_loop)

        except RuntimeError:
            _glob_loop = asyncio.new_event_loop()
            setGreedCoro(_glob_loop)

            _glob_thrd = threading.Thread(target=_glob_loop.run_forever, name='SynLoop')
            _glob_thrd.setDaemon(True)
            _glob_thrd.start()

    return _glob_loop

def setGreedCoro(loop: asyncio.AbstractEventLoop):
    greedy_threshold = os.environ.get('SYN_GREEDY_CORO')
    if greedy_threshold is not None:  # pragma: no cover
        logger.info(f'Setting ioloop.slow_callback_duration to {greedy_threshold}')
        loop.set_debug(True)
        loop.slow_callback_duration = float(greedy_threshold)

def iAmLoop():
    initloop()
    return threading.currentThread() == _glob_thrd

def sync(coro, timeout=None):
    '''
    Schedule a coroutine to run on the global loop and return it's result.

    Args:
        coro (coroutine): The coroutine instance.

    Notes:
        This API is thread safe and should only be called by non-loop threads.
    '''
    loop = initloop()
    return asyncio.run_coroutine_threadsafe(coro, loop).result(timeout)

def synchelp(f):
    '''
    The synchelp decorator allows the transparent execution of
    a coroutine using the global loop from a thread other than
    the event loop.  In both use cases, the actual work is done
    by the global event loop.

    Examples:

        Use as a decorator::

            @s_glob.synchelp
            async def stuff(x, y):
                await dostuff()

        Calling the stuff function as regular async code using the standard await syntax::

            valu = await stuff(x, y)

        Calling the stuff function as regular sync code outside of the event loop thread::

            valu = stuff(x, y)

    '''
    def wrap(*args, **kwargs):

        coro = f(*args, **kwargs)

        if not iAmLoop():
            return sync(coro)

        return coro

    return wrap



import asyncio

import synapse.lib.boss as s_boss
import synapse.tests.utils as s_test

class BossTest(s_test.SynTest):

    async def test_boss_base(self):

        async with await s_boss.Boss.anit() as boss:

            evnt = asyncio.Event()

            async def testloop():
                evnt.set()
                while True:
                    await asyncio.sleep(10)

            self.len(0, boss.ps())

            synt = await boss.promote('test', None, info={'hehe': 'haha'})

            self.len(1, boss.ps())

            self.eq('test', synt.name)
            self.eq('haha', synt.info.get('hehe'))

            synt0 = await boss.execute(testloop(), 'testloop', None, info={'foo': 'bar'})

            await evnt.wait()

            self.len(2, boss.ps())

            await synt0.kill()

            self.len(1, boss.ps())



import os
import asyncio

import synapse.common as s_common

import synapse.lib.cmdr as s_cmdr
import synapse.lib.encoding as s_encoding
import synapse.lib.lmdbslab as s_lmdbslab

import synapse.tests.utils as s_t_utils

from synapse.tests.utils import alist


class CmdCoreTest(s_t_utils.SynTest):

    async def test_storm(self):

        help_msg = 'Execute a storm query.'

        async with self.getTestCoreAndProxy() as (realcore, core):

            await realcore.addTagProp('score', ('int', {}), {})

            await self.agenlen(1, core.eval("[ test:str=abcd :tick=2015 +#cool]"))

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('help storm')
            outp.expect(help_msg)

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm help')
            outp.expect('For detailed help on any command')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm')
            outp.expect(help_msg)

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --debug test:str=abcd')
            outp.expect("('init',")
            outp.expect("('node',")
            outp.expect("('fini',")
            outp.expect("tick")
            outp.expect("tock")
            outp.expect("took")

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --debug test:str=zzz')
            outp.expect("('init',")
            self.false(outp.expect("('node',", throw=False))
            outp.expect("('fini',")
            outp.expect("tick")
            outp.expect("tock")
            outp.expect("took")

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm test:str=b')
            outp.expect('complete. 0 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm test:str=abcd')
            outp.expect(':tick = 2015/01/01 00:00:00.000')
            outp.expect('#cool')
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --show-nodeedits [test:int=42]')
            outp.expect('node:edits')
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --editformat count [test:int=43]')
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --hide-tags test:str=abcd')
            outp.expect(':tick = 2015/01/01 00:00:00.000')
            self.false(outp.expect('#cool', throw=False))
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --hide-props test:str=abcd')
            self.false(outp.expect(':tick = 2015/01/01 00:00:00.000', throw=False))
            outp.expect('#cool')
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --show print,foo:bar test:str=abcd')
            self.false(outp.expect(':tick = 2015/01/01 00:00:00.000', throw=False))
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --hide-tags --hide-props test:str=abcd')
            self.false(outp.expect(':tick = 2015/01/01 00:00:00.000', throw=False))
            self.false(outp.expect('#cool', throw=False))
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --raw test:str=abcd')
            outp.expect("'tick': 1420070400000")
            outp.expect("'tags': {'cool': (None, None)")
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --bad')
            outp.expect('Syntax Error')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm newpz')
            outp.expect('NoSuchName')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --hide-unknown [test:str=1234]')
            s = str(outp)
            self.notin('node:add', s)
            self.notin('prop:set', s)
            await self.agenlen(1, core.eval('[test:comp=(1234, 5678)]'))

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            q = 'storm --raw --path test:comp -> test:int'
            await cmdr.runCmdLine(q)
            self.true(outp.expect("('test:int', 1234)"))
            self.true(outp.expect("'path'"))

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm --show-prov [ test:str=foo +#bar.baz=(2015,?) ]')
            self.true(outp.expect('prov:new'))
            self.true(outp.expect('....\ntest:str'))
            self.true(outp.expect('#bar.baz = (2015/01/01 00:00:00.000, ?)', throw=False))
            self.false(outp.expect('#bar ', throw=False))
            outp.expect('complete. 1 nodes')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm [ test:str=foo +#bar:score=22 +#bar.baz:score=0 ]')
            self.true(outp.expect('#bar:score = 22', throw=False))
            self.true(outp.expect('#bar.baz = (2015/01/01 00:00:00.000, ?)', throw=False))
            self.true(outp.expect('#bar.baz:score = 0', throw=False))
            self.false(outp.expect('#bar ', throw=False))
            outp.expect('complete. 1 nodes')

            # Warning test
            guid = s_common.guid()
            await alist(core.eval(f'[test:guid={guid}]'))
            await alist(core.eval(f'[test:edge=(("test:guid", {guid}), ("test:str", abcd))]'))

            q = 'storm test:str=abcd <- test:edge :n1:form -> *'
            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine(q)
            e = 'WARNING: The source property "n1:form" type "str" is not a form. Cannot pivot.'
            self.true(outp.expect(e))

            # Err case
            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine('storm test:str -> test:newp')
            self.true(outp.expect('ERROR'))
            self.true(outp.expect('NoSuchProp'))
            self.true(outp.expect('test:newp'))

            # Cancelled case
            evnt = asyncio.Event()
            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)

            def setEvt(event):
                smsg = event[1].get('mesg')
                if smsg[0] == 'node':
                    evnt.set()

            async def runLongStorm():
                with cmdr.onWith('storm:mesg', setEvt):
                    await cmdr.runCmdLine('storm .created | sleep 10')

            task = realcore.schedCoro(runLongStorm())
            self.true(await asyncio.wait_for(evnt.wait(), timeout=6))
            ps = await core.ps()
            self.len(1, ps)
            iden = ps[0].get('iden')
            await core.kill(iden)
            await asyncio.sleep(0)
            self.true(outp.expect('query canceled.'))
            self.true(task.done())

            # Color test
            outp.clear()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
            await cmdr.runCmdLine(f'storm test:{"x"*50} -> * -> $')
            outp.expect('-> *')
            outp.expect('Syntax Error')

            outp.clear()
            with self.withCliPromptMock() as patch:
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                cmdr.colorsenabled = True
                await cmdr.runCmdLine('storm [#foo]')
                await cmdr.runCmdLine('storm test:str ->')
            lines = self.getMagicPromptColors(patch)
            clines = []
            for (color, text) in lines:
                if text.startswith('Syntax Error:'):
                    text = 'Syntax Error'
                clines.append((color, text))
            self.isin(('#6faef2', '[#foo]'), clines)
            self.isin(('#6faef2', ' ^'), clines)
            self.isin(('#ff0066', 'Syntax Error'), clines)
            self.isin(('#6faef2', 'test:str ->'), clines)
            self.isin(('#6faef2', '           ^'), clines)

            # Trying to print an \r doesn't assert (prompt_toolkit bug)
            # https://github.com/prompt-toolkit/python-prompt-toolkit/issues/915
            await core.addNode('test:str', 'foo', props={'hehe': 'windows\r\nwindows\r\n'})
            await cmdr.runCmdLine('storm test:str=foo')
            self.true(1)

            await realcore.nodes('[ inet:ipv4=1.2.3.4 +#visi.woot ]')
            await s_lmdbslab.Slab.syncLoopOnce()

    async def test_log(self):

        def check_locs_cleanup(cobj):
            keys = list(cobj.locs.keys())
            for key in keys:
                if key.startswith('log:'):
                    self.fail(f'Key with "log:" prefix found. [{key}]')

        async with self.getTestCoreAndProxy() as (realcore, core):

            with self.getTestSynDir() as dirn:

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                await cmdr.runCmdLine('log --on --format jsonl')
                fp = cmdr.locs.get('log:fp')
                await cmdr.runCmdLine('storm --editformat splices [test:str=hi :tick=2018 +#haha.hehe]')

                await cmdr.runCmdLine('storm --editformat nodeedits [test:str=hi2 :tick=2018 +#haha.hehe]')
                await cmdr.runCmdLine('storm [test:comp=(42, bar)]')

                # Try calling on a second time - this has no effect on the
                # state of cmdr, but prints a warning
                await cmdr.runCmdLine('log --on --format jsonl')

                await cmdr.runCmdLine('log --off')
                await cmdr.fini()
                check_locs_cleanup(cmdr)

                self.true(outp.expect('Starting logfile'))
                e = 'Must call --off to disable current file before starting a new file.'
                self.true(outp.expect(e))
                self.true(outp.expect('Closing logfile'))
                self.true(os.path.isfile(fp))

                # Ensure that jsonl is how the data was saved
                with s_common.genfile(fp) as fd:
                    genr = s_encoding.iterdata(fd, close_fd=False, format='jsonl')
                    objs = list(genr)
                self.eq(objs[0][0], 'init')

                nodeedits = [m for m in objs if m[0] == 'node:edits']
                self.ge(len(nodeedits), 2)

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                # Our default format is mpk
                fp = os.path.join(dirn, 'loggyMcLogFace.mpk')
                await cmdr.runCmdLine(f'log --on --edits-only --path {fp}')
                fp = cmdr.locs.get('log:fp')
                await cmdr.runCmdLine('storm [test:str="I am a message!" :tick=1999 +#oh.my] ')
                await cmdr.runCmdLine('log --off')
                await cmdr.fini()
                check_locs_cleanup(cmdr)

                self.true(os.path.isfile(fp))
                with s_common.genfile(fp) as fd:
                    genr = s_encoding.iterdata(fd, close_fd=False, format='mpk')
                    objs = list(genr)

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                # Our default format is mpk
                fp = os.path.join(dirn, 'loggyMcNodeFace.mpk')
                await cmdr.runCmdLine(f'log --on --nodes-only --path {fp}')
                fp = cmdr.locs.get('log:fp')
                await cmdr.runCmdLine('storm [test:str="I am a message!" :tick=1999 +#oh.my] ')
                await cmdr.runCmdLine('log --off')
                await cmdr.fini()
                check_locs_cleanup(cmdr)

                self.true(os.path.isfile(fp))
                with s_common.genfile(fp) as fd:
                    genr = s_encoding.iterdata(fd, close_fd=False, format='mpk')
                    objs = list(genr)
                self.eq(objs[0][0], ('test:str', 'I am a message!'))

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                await cmdr.runCmdLine('log --on --off')
                await cmdr.fini()
                self.true(outp.expect('log: error: argument --off: not allowed with argument --on'))

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                await cmdr.runCmdLine('log')
                await cmdr.fini()
                self.true(outp.expect('log: error: one of the arguments --on --off is required'))

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                await cmdr.runCmdLine('log --on --edits-only --nodes-only')
                await cmdr.fini()
                e = 'log: error: argument --nodes-only: not allowed with argument --edits-only'
                self.true(outp.expect(e))

                # Bad internal state
                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)
                await cmdr.runCmdLine('log --on --nodes-only')
                cmdr.locs['log:fmt'] = 'newp'
                with self.getAsyncLoggerStream('synapse.cmds.cortex',
                                               'Unknown encoding format: newp') as stream:
                    await cmdr.runCmdLine('storm test:str')
                    self.true(await stream.wait(2))

                await cmdr.fini()

    async def test_storm_save_nodes(self):

        async with self.getTestCoreAndProxy() as (core, prox):

            dirn = s_common.gendir(core.dirn, 'junk')
            path = os.path.join(dirn, 'nodes.jsonl')

            await core.nodes('[ test:int=20 test:int=30 ]')

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(prox, outp=outp)
            await cmdr.runCmdLine(f'storm --save-nodes {path} test:int')
            outp.expect('2 nodes')

            jsdata = [item for item in s_common.jslines(path)]
            self.len(2, jsdata)
            self.eq({tuple(n[0]) for n in jsdata},
                    {('test:int', 20), ('test:int', 30)})

    async def test_storm_file_optfile(self):

        async with self.getTestCoreAndProxy() as (core, prox):

            test_opts = {'vars': {'hehe': 'woot.com'}}
            dirn = s_common.gendir(core.dirn, 'junk')

            optsfile = os.path.join(dirn, 'woot.json')
            optsfile_yaml = os.path.join(dirn, 'woot.yaml')
            stormfile = os.path.join(dirn, 'woot.storm')

            with s_common.genfile(stormfile) as fd:
                fd.write(b'[ inet:fqdn=$hehe ]')

            s_common.jssave(test_opts, optsfile)
            s_common.yamlsave(test_opts, optsfile_yaml)

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(prox, outp=outp)
            await cmdr.runCmdLine(f'storm --optsfile {optsfile} --file {stormfile}')
            self.true(outp.expect('inet:fqdn=woot.com'))

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(prox, outp=outp)
            await cmdr.runCmdLine(f'storm --optsfile {optsfile_yaml} --file {stormfile}')
            self.true(outp.expect('inet:fqdn=woot.com'))

            # Sad path cases
            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(prox, outp=outp)
            await cmdr.runCmdLine(f'storm --file {stormfile} --optsfile {optsfile} .created')
            self.true(outp.expect('Cannot use a storm file and manual query together.'))
            self.false(outp.expect('inet:fqdn=woot.com', throw=False))

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(prox, outp=outp)
            await cmdr.runCmdLine(f'storm --file {stormfile} --optsfile newp')
            self.true(outp.expect('optsfile not found'))

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(prox, outp=outp)
            await cmdr.runCmdLine(f'storm --file newp --optsfile {optsfile}')
            self.true(outp.expect('file not found'))



import synapse.common as s_common
import synapse.tests.utils as s_t_utils

class CnGovTest(s_t_utils.SynTest):

    async def test_models_cngov_mucd(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:
                org0 = s_common.guid()
                props = {
                    'org': org0
                }
                node = await snap.addNode('gov:cn:icp', 12345678, props)
                self.eq(node.ndef[1], 12345678)
                self.eq(node.get('org'), org0)

                node = await snap.addNode('gov:cn:mucd', 61786)
                self.eq(node.ndef[1], 61786)
